# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from collections import namedtuple
import six
from boundary_layer.plugins import util
from boundary_layer.logger import logger


OoziePluginContainer = namedtuple('OoziePluginContainer', ['name', 'priority', 'plugin'])


class OozieMetaPlugin(object):
    def __init__(self, plugins, args):
        self.args = args

        sorted_plugins = sorted(
            plugins,
            key=lambda p: p.priority.value,
            reverse=True)

        self._plugin_containers = [
            OoziePluginContainer(
                name=plugin.name,
                priority=plugin.priority,
                plugin=plugin.oozie_plugin_cls(args)) for plugin in sorted_plugins]

    def action_builders(self):
        return util.merge_lists(
            pc.plugin.action_builders() for pc in self._plugin_containers)

    def plugin_config(self):
        all_configs = {
            plugin_container.name: plugin_container.plugin.plugin_config()
            for plugin_container in self._plugin_containers
        }

        return {name: config for (name, config) in six.iteritems(all_configs) if config}

    def dag_args(self):
        return util.merge_dicts(
            pc.plugin.dag_args() for pc in self._plugin_containers)

    def default_task_args(self):
        return util.merge_dicts(
            pc.plugin.default_task_args() for pc in self._plugin_containers)

    def dag_imports(self):
        return util.merge_dicts(
            pc.plugin.dag_imports() for pc in self._plugin_containers)

    def cluster_config(self):
        plugins_with_cluster_configs = [
            pc for pc in self._plugin_containers
            if pc.plugin.cluster_config()]

        if not plugins_with_cluster_configs:
            raise Exception('No cluster configurations found for oozie parser!')

        if len(plugins_with_cluster_configs) > 1:
            logger.info(
                'Multiple cluster configurations found.  Choosing configuration '
                'from plugin `%s`, with priority `%s`',
                plugins_with_cluster_configs[0].name,
                plugins_with_cluster_configs[0].priority)

        return plugins_with_cluster_configs[0].plugin.cluster_config()

    def upstream_operators(self):
        return util.merge_lists(
            pc.plugin.upstream_operators() for pc in self._plugin_containers)

    def jsp_macros(self):
        return util.merge_dicts(
            pc.plugin.jsp_macros() for pc in self._plugin_containers)


class BaseOozieParserPlugin(object):
    def action_builders(self):
        """ Plugin may provide customized parsers for oozie actions.
            Must be a list of OozieActionBuilder classes
        """
        return []

    def plugin_config(self):
        return {}

    def dag_args(self):
        return {}

    def default_task_args(self):
        return {}

    def cluster_config(self):
        return None

    def dag_imports(self):
        return {}

    def upstream_operators(self):
        return []

    def jsp_macros(self):
        return {}

    @classmethod
    def register_arguments(cls, parser):
        # Overrideable method for adding command-line arguments
        pass

    def __init__(self, args):
        self.args = args



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re
import datetime
from collections import namedtuple
import json
import base64
import marshmallow as ma
import jinja2
from boundary_layer.registry.types.preprocessor import PropertyPreprocessor
from boundary_layer.schemas.base import StrictSchema
from boundary_layer.logger import logger
from boundary_layer.util import GenericNamedParameterPasser


class DateStringToDatetime(PropertyPreprocessor):
    type = "date_string_to_datetime"

    def imports(self):
        return {'modules': ['datetime']}

    def process_arg(self, arg, node, raw_args):
        date = None
        try:
            date = datetime.datetime.strptime(arg, '%Y-%m-%d')
        except ValueError as e:
            raise Exception(
                'Error in preprocessor {} for argument `{}`: {}'.format(
                    self.type,
                    arg,
                    str(e)))

        return date


class BuildKubernetesSchema(StrictSchema):
    class_name = ma.fields.String(required=True)

    @ma.validates_schema
    def check_valid_class(self, data):
        ALLOWED_CLASS = ['airflow.contrib.kubernetes.volume.Volume',
                         'airflow.contrib.kubernetes.volume_mount.VolumeMount',
                         'airflow.contrib.kubernetes.secret.Secret',
                         'airflow.contrib.kubernetes.pod.Resources']
        if data.get('class_name') not in ALLOWED_CLASS:
            raise ma.ValidationError(
                '`class_name` must be one of `{}`'.format(
                    '`, `'.join(ALLOWED_CLASS)),
                ['class_name'])


class KubernetesPrep(PropertyPreprocessor):
    type = "kubernetes_prep"
    properties_schema_cls = BuildKubernetesSchema

    def imports(self):
        return {'modules': ['.'.join(self.properties['class_name'].split('.')[:-1])]}

    def process_arg(self, arg, node, raw_args):
        kube_objects = None
        try:
            if isinstance(arg, list):
                kube_objects = [GenericNamedParameterPasser(self.properties['class_name'], a)
                                for a in arg]
            else:
                kube_objects = GenericNamedParameterPasser(self.properties['class_name'], arg)
        except Exception as e:
            raise Exception(
                'Error in preprocessor {} for argument `{}`: {}'.format(
                    self.type,
                    arg,
                    str(e)))

        return kube_objects


class BuildTimedeltaSchema(StrictSchema):
    units = ma.fields.String(required=True)

    @ma.validates_schema
    def check_valid_units(self, data):
        ALLOWED_UNITS = ['seconds', 'minutes', 'hours', 'days']
        if data.get('units') not in ALLOWED_UNITS:
            raise ma.ValidationError(
                '`units` must be one of `{}`'.format(
                    '`, `'.join(ALLOWED_UNITS)),
                ['units'])


class BuildTimedelta(PropertyPreprocessor):
    type = "to_timedelta"

    properties_schema_cls = BuildTimedeltaSchema

    def imports(self):
        return {'modules': ['datetime']}

    def process_arg(self, arg, node, raw_args):
        delta = None

        try:
            timedelta_arg = {self.properties['units']: arg}
            delta = datetime.timedelta(**timedelta_arg)
        except TypeError as e:
            raise Exception(
                'Error in preprocessor {} for argument `{}`: {}'.format(
                    self.type,
                    arg,
                    str(e)))

        return delta


class EnsureRenderedStringPatternSchema(StrictSchema):
    pattern = ma.fields.String(required=True)


class EnsureRenderedStringPattern(PropertyPreprocessor):
    type = "ensure_rendered_string_pattern"

    properties_schema_cls = EnsureRenderedStringPatternSchema

    def process_arg(self, arg, node, raw_args):
        regex = None
        try:
            regex = re.compile(self.pattern)
        except Exception:
            raise Exception(
                'Error compiling regex for `{}`: `{}` is an invalid pattern'.format(
                    self.type,
                    self.properties['pattern']))

        rendered_arg = None
        try:
            rendered_arg = self.render_template(arg, raw_args)
        except jinja2.exceptions.UndefinedError:
            logger.warning(
                'Could not render template `%s`; cannot verify that the argument '
                'matches the required pattern `%s`!',
                arg,
                regex.pattern)
            return arg

        if regex.match(rendered_arg):
            # return the original arg, not the rendered arg, because we are not
            # actually transforming anything, just validating
            return arg

        VERBATIM_REGEX = '<<.+>>'
        if re.compile(VERBATIM_REGEX).search(rendered_arg):
            logger.warning(
                'Argument generated from `%s` may not match the required pattern `%s` and fail.',
                rendered_arg,
                regex.pattern)
            return arg

        raise Exception(
            'Invalid argument `{}`: does not match expected pattern `{}`'.format(
                rendered_arg,
                regex.pattern))

    @property
    def pattern(self):
        # Return the pattern configured in the properties, but make sure that
        # the pattern ends in a `$` to ensure that it applies to the entire
        # string, not a prefix of it

        pattern = self.properties['pattern'].rstrip('$')

        return pattern + '$'

    def render_template(self, arg, raw_args):
        env = jinja2.Environment(undefined=jinja2.StrictUndefined)

        return env.from_string(arg).render(**self._template_context(raw_args))

    @staticmethod
    def _template_context(raw_args):
        """ Create an emulated jinja context for common airflow date-related
            macros. The actual values of the macros do not matter, we are only
            concerned with general formatting which will be properly represented
            for any date.

            :returns: the context
            :rtype: dict
        """

        now = datetime.datetime.utcnow()

        context = {
            'ts': now.isoformat(),
            'ts_nodash': now.isoformat().replace('-', ''),
        }

        TaskTuple = namedtuple('TaskTuple', ['task_id'])
        task_id = raw_args.get('task_id')
        if task_id:
            context['task'] = TaskTuple(task_id=task_id)

        for field_name in ['ds', 'yesterday_ds', 'tomorrow_ds', 'next_ds', 'prev_ds']:
            context[field_name] = now.strftime('%Y-%m-%d')
            context[field_name + '_nodash'] = now.strftime('%Y%m%d')

        for field_name in ['execution_date', 'prev_execution_date', 'next_execution_date']:
            context[field_name] = now

        return context

class PubsubMessageDataToBinaryString(PropertyPreprocessor):
    """
    Converts pubsub message data with various python types 
    to binary strings.
    Supported message data arg types: `dict`, and `str`

    If arg is type `dict` that data will be converted
    into a json string, and then encoded into a binary string

    This preprocessing is necessary as messages with the `data`
    property are required to be binary strings to use airflow's
    provided pubsub functionality
    """
    type = "pubsub_message_data_to_binary_string"

    def imports(self):
        return {'modules': ['json', 'base64']}

    def process_arg(self, arg, node, raw_args):
        """
        Given array of messages, if `data` property is present in
        message objects, convert to binary string
        """
        res = []
        for message in arg:
            if 'data' in message:
                message['data'] = self._process_data_arg(message['data'])
            res.append(message)
        return res

    def _process_data_arg(self, arg):
        """
        Given either a dict or str:
        - If str, encode to binary string
        - If dict, convert to json and encode as binary string
        """
        bin_string = None
        try:
            res_str = arg if not self._is_dict(arg) else self._json_handler(arg)
            bin_string = base64.b64encode(res_str.encode('utf-8'))
        except Exception as e:
            raise Exception(
                'Error in preprocessor {} for argument `{}`: {}'.format(
                    self.type,
                    arg,
                    str(e)))
        return bin_string

    def _json_handler(self, arg):
        return json.dumps(arg)

    def _is_dict(self, arg):
        return isinstance(arg, dict)



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import datetime
import six
from six.moves import filter, map
from boundary_layer.util import sanitize_operator_name
from boundary_layer.util import GenericNamedParameterPasser


def verbatim(item):
    if isinstance(item, list):
        return list(map(verbatim, item))
    elif isinstance(item, six.string_types):
        return '<<' + item + '>>'

    raise Exception('Cannot cast item `{}` to a verbatim string!'.format(
        item))


def construct_dag_name(reference_path):
    return '.'.join(map(sanitize_operator_name, reference_path))


def subdag_builder_name(sub_workflow_name):
    return '{}_builder'.format(sanitize_operator_name(sub_workflow_name))


def add_leading_spaces(text_block, indent_depth):
    indent_spaces = 4
    num_spaces = indent_spaces * indent_depth
    return '\n'.join((' ' * num_spaces + line) for line in text_block.split('\n'))


def enquote(x):
    if "'" not in x:
        return "'%s'" % x

    if '"' not in x:
        return '"%s"' % x

    assert '"""' not in x, 'Cannot wrap quotes around string: {}'.format(x)

    return '"""%s"""' % x


def split_verbatim(value):
    """ Method for handling verbatim indicator characters in strings, which
        are used to indicate that the string should be inserted verbatim into
        the dag as python code, rather than as a string. These are used,
        for example, to allow us to write simple templates for the "generator"
        actions, which construct a for-loop and have to refer to the loop
        variable.  We use the delimiters `<<` and `>>` to indicate verbatim
        strings.  I hope nobody wants to use the left- and right-shift
        inside one of these things...
    """
    def wrap(item, in_block):
        """ non-verbatim strings must be wrapped in quotes; we wrap verbatim
            strings in parentheses in order to ensure proper order-of-operations
            when we join together the components with '+'
        """
        if not item:
            return ''

        return enquote(item) if not in_block else '({})'.format(item)

    def _do_split(remaining, in_block):
        if not in_block and '<<' not in remaining:
            return [wrap(remaining, in_block)]

        if in_block and '>>' not in remaining:
            raise Exception(
                'Unpaired verbatim string character `<<` in string {}'.format(value))

        idx_start = remaining.index('<<' if not in_block else '>>')

        if in_block and '<<' in remaining and remaining.index('<<') < idx_start:
            raise Exception(
                'Invalid nested verbatim string characters in string {}'.format(value))

        before_slice = remaining[:idx_start]
        after_slice = remaining[(2+idx_start):]

        return [wrap(before_slice, in_block)] + \
            _do_split(after_slice, not in_block)

    return list(filter(None, _do_split(remaining=value, in_block=False)))


def format_value(value):
    """ Format values for passing arguments to airflow operators.  This
        involves differentiating between string components that should be
        inserted verbatim (e.g. variable references) from those that should be
        quoted.  Our convention for verbatim strings is that they are inclosed in
        delimiters << >>.  For example, the following string:
            "<< '10' if some_value else '20' >> dags"
        will be transformed to the python code:
            ('10' if some_value else '20') + 'dags'
        This enables the yaml dags specifications to contain arbitrary
        logical statements.
    """
    if isinstance(value, list):
        return '[{}]'.format(','.join(map(format_value, value)))

    if isinstance(value, dict):
        pairs = ['{}: {}'.format(enquote(k), format_value(v))
                 for (k, v) in value.items()]

        return '{{ {items} }}'.format(items=','.join(pairs))

    if isinstance(value, (int, float, type(None))):
        return str(value)

    if isinstance(value, (datetime.datetime, datetime.timedelta, GenericNamedParameterPasser)):
        return format_value('<<{}>>'.format(repr(value)))

    if not isinstance(value, six.string_types):
        raise Exception('Cannot format value `{}`: no handler for type {}'.format(
            str(value),
            type(value)))

    components = split_verbatim(value)

    if len(components) == 1:
        return components[0]

    return '({})'.format(' + '.join(components))


def comment(value):
    lines = value.split('\n')
    commented = ['# ' + line for line in lines]
    return '\n'.join(commented)



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import abc
from boundary_layer.registry import Registry, RegistryNode, NodeTypes


class PropertyPreprocessor(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def process_arg(self, arg, node, raw_args):
        pass

    @abc.abstractproperty
    def type(self):
        pass

    def imports(self):
        return {}

    def __init__(self, properties):
        if not self.properties_schema_cls and properties:
            raise Exception(
                'Invalid properties `{}`: preprocessor `{}` does not '
                'take any properties!'.format(
                    properties,
                    self.type))

        if not self.properties_schema_cls:
            self.properties = None
            return

        loaded = self.properties_schema_cls().load(properties)
        if loaded.errors:
            raise Exception(
                'Error parsing properties for preprocessor `{}`: {}'.format(
                    self.type,
                    loaded.errors))

        self.properties = loaded.data

    @property
    def properties_schema_cls(self):
        """ Any preprocessor that requires a configuration must define a marshmallow
            schema for that configuration.  If no schema is provided, then
            we will reject any properties that are provided.
        """
        pass


class PropertyPreprocessorNode(RegistryNode):
    type = NodeTypes.PREPROCESSOR

    @property
    def configured_preprocessor(self):
        return self.config(self.item.get('properties', {}))

    def process_arg(self, arg, node, raw_args):
        return self.configured_preprocessor.process_arg(arg, node, raw_args)

    def imports(self):
        return self.configured_preprocessor.imports()


class PropertyPreprocessorRegistry(Registry):
    node_cls = PropertyPreprocessorNode

    @property
    def node_configs(self):
        return self._node_configs

    def __init__(self, preprocessor_classes):
        self._node_configs = {
            pp_cls.type: pp_cls for pp_cls in preprocessor_classes
        }



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re
from datetime import timedelta
import semver
from marshmallow import fields, validates_schema, ValidationError, post_dump
from boundary_layer import VERSION, MIN_SUPPORTED_VERSION
from boundary_layer.schemas.base import StrictSchema


class OperatorSchema(StrictSchema):
    name = fields.String(required=True)
    type = fields.String(required=True)

    requires_resources = fields.List(fields.String())

    upstream_dependencies = fields.List(fields.String())
    downstream_dependencies = fields.List(fields.String())

    properties = fields.Dict()


class ReferenceSchema(OperatorSchema):
    target = fields.String(required=True)


class GeneratorSchema(ReferenceSchema):
    auto_task_id_mode = fields.String()
    regex_blocklist = fields.List(fields.String())

    @validates_schema
    def check_task_id_mode(self, data):
        if 'auto_task_id_mode' not in data:
            return

        allowed_values = ('index', 'item_name')
        if data['auto_task_id_mode'] not in allowed_values:
            raise ValidationError(
                'Value must be one of {}'.format(allowed_values),
                ['auto_task_id_mode'])


class ResourceSchema(OperatorSchema):
    # Resources are like operators, but they must always have both a
    # create step and a destroy step.  An example is a transient cluster
    # on DataProc / EMR.  These will generally be specified
    # by the class referenced by the `type` argument, although we
    # should also permit a 'custom' type which specifies create/destroy
    # arguments in the config directly

    # An additional optional parameter allows the DAG-writer to disable
    # the creation of sentinel nodes for cases in which the destroy
    # operator is a leaf node.
    disable_sentinel_node = fields.Boolean()


class ObjectImportSchema(StrictSchema):
    module = fields.String(required=True)
    objects = fields.List(fields.String(), required=True)


class ImportSchema(StrictSchema):
    modules = fields.List(fields.String())
    objects = fields.List(fields.Nested(ObjectImportSchema))


class BaseDagSchema(StrictSchema):
    name = fields.String(required=True)

    imports = fields.Nested(ImportSchema)

    resources = fields.List(fields.Nested(ResourceSchema))

    before = fields.List(fields.Nested(OperatorSchema))
    operators = fields.List(fields.Nested(OperatorSchema))
    after = fields.List(fields.Nested(OperatorSchema))

    # we treat sub-dags and generators as a special first-class category of
    # Reference operators.
    sub_dags = fields.List(fields.Nested(ReferenceSchema()))
    generators = fields.List(fields.Nested(GeneratorSchema()))

    plugin_config = fields.Dict()

    @validates_schema
    def validate_plugin_config(self, data):
        from boundary_layer import plugins
        if 'plugin_config' not in data:
            return

        try:
            plugins.manager.validate_config(data['plugin_config'])
        except ValidationError as e:
            raise e
        except Exception as e:
            raise ValidationError(str(e))


class DagArgsSchema(StrictSchema):
    catchup = fields.Boolean(missing=True)
    max_active_runs = fields.Integer(missing=1)
    concurrency = fields.Integer()
    # schedule_interval argument supports cron strings (e.g. 0 * * * *),
    # '@hourly/daily/etc', or numeric (seconds)
    schedule_interval = fields.String(allow_none=True)
    description = fields.String()
    dagrun_timeout = fields.Integer()
    default_view = fields.String()
    orientation = fields.String()
    sla_miss_callback = fields.String()
    on_success_callback = fields.String()
    on_failure_callback = fields.String()
    params = fields.Dict()
    full_filepath = fields.String()
    template_searchpath = fields.List(fields.String())
    template_undefined = fields.String()
    user_defined_macros = fields.Dict()
    user_defined_filters = fields.Dict()
    doc_md = fields.String()
    access_control = fields.Dict()

    @validates_schema
    def validate_callbacks(self, data):
        callbacks = ['sla_miss_callback', 'on_success_callback',
                     'on_failure_callback']
        for cb in callbacks:
            if cb not in data:
                continue
            if not re.compile('<<.+>>').match(data[cb]):
                raise ValidationError(
                    '{} must be a verbatim string like <<...>>'.format(cb),
                    [cb])

    @validates_schema
    def validate_default_view(self, data):
        if 'default_view' not in data:
            return

        allowed_values = ('tree', 'graph', 'duration', 'gantt',
                          'landing_times')
        if data['default_view'] not in allowed_values:
            raise ValidationError(
                'Value must be one of {}'.format(allowed_values),
                ['default_view'])

    @validates_schema
    def validate_orientation(self, data):
        if 'orientation' not in data:
            return

        allowed_values = ('LR', 'TB', 'RL', 'BT')
        if data['orientation'] not in allowed_values:
            raise ValidationError(
                'Value must be one of {}'.format(allowed_values),
                ['orientation'])

    @validates_schema
    def validate_template_undefined(self, data):
        if 'template_undefined' not in data:
            return
        if not re.compile('<<.+>>').match(data['template_undefined']):
            raise ValidationError(
                'template_undefined must be a verbatim string like <<...>>',
                ['template_undefined'])

    @post_dump
    def dagrun_timeout_to_timedelta(self, data):
        if not self.context.get('for_dag_output'):
            return data
        if 'dagrun_timeout' in data:
            try:
                delta = timedelta(seconds=data['dagrun_timeout'])
            except TypeError as e:
                raise Exception(
                    'Error in making dagrun_timeout into timedelta object : {}'
                    .format(str(e)))
            data['dagrun_timeout'] = delta
        return data


class PrimaryDagSchema(BaseDagSchema):
    compatibility_version = fields.String()

    dag_args = fields.Nested(DagArgsSchema)

    default_task_args = fields.Dict()

    @validates_schema
    def validate_compatibility_version(self, data):
        if not data.get('compatibility_version'):
            return

        version = None
        try:
            version = semver.parse_version_info(data['compatibility_version'])
        except ValueError:
            raise ValidationError('Must be a valid SemVer',
                                  ['compatibility_version'])

        if VERSION < version:
            raise ValidationError(
                'Incompatible boundary_layer version: This '
                'workflow requires boundary_layer version {} or higher! '
                'Current version is {}'.format(version, VERSION),
                ['compatibility_version'])

        if version < MIN_SUPPORTED_VERSION:
            raise ValidationError(
                'Incompatible boundary_layer version: This workflow '
                'is for the incompatible prior version {}. Use the '
                'migrate-workflow script to update it.'.format(version),
                ['compatibility_version'])



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

# Simple utilities for writing out YAML configs in a more readable format: in particular,
# with singleton values appearing before arrays, shorter arrays before longer arrays,
# and certain higher-priority keys appearing before everything else.
# Based on an OrderedDict-representer approach described here:
# https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
import functools
from collections import OrderedDict
import yaml
from six.moves import map


def _build_dumper():
    class OrderedDumper(yaml.SafeDumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            data.items())

    OrderedDumper.add_representer(OrderedDict, _dict_representer)

    return OrderedDumper


PRIORITY_KEYS = [
    'name',
    'type',
    'compatibility_version',
    'plugin_config',
    'imports',
    'dag_args',
    'resources',
    'default_task_args',
    'before',
    'operators',
    'after',
    'sub_dags',
    'generators']


def _reorder(item):
    if not isinstance(item, dict):
        return item

    def scorer(key, value):
        if key in PRIORITY_KEYS:
            return PRIORITY_KEYS.index(key) - len(PRIORITY_KEYS)

        value_score = len(value) if isinstance(value, (dict, list)) else 0

        return value_score

    def comparator(kv0, kv1):
        score0 = scorer(*kv0)
        score1 = scorer(*kv1)

        if score0 != score1:
            return 1 if score0 > score1 else -1
        else:
            (key0, _) = kv0
            (key1, _) = kv1

            if key0 == key1:
                return 0

            return 1 if key0 > key1 else -1

    result = OrderedDict()
    for (key, value) in sorted(item.items(), key=functools.cmp_to_key(comparator)):
        if isinstance(value, dict):
            result[key] = _reorder(value)
            continue

        if isinstance(value, list):
            result[key] = list(map(_reorder, value))
            continue

        result[key] = value

    return result


def dump_all(items):
    return yaml.dump_all(
        list(map(_reorder, items)),
        None,
        _build_dumper(),
        default_flow_style=False)


def dump(item):
    return dump_all([item])



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import marshmallow as ma


class OozieBaseSchema(ma.Schema):
    singletons_to_lists = []

    @ma.pre_load
    def _convert_singletons_to_lists(self, data):
        if all(isinstance(data.get(key, []), list) for key in self.singletons_to_lists):
            return data

        copy = data.copy()

        for key in self.singletons_to_lists:
            if not isinstance(data.get(key, []), list):
                copy[key] = [data.get(key)]

        return copy


class OozieNamedObjectSchema(OozieBaseSchema):
    name = ma.fields.String(required=True, load_from='@name')


class OozieFlowControlSchema(OozieBaseSchema):
    to = ma.fields.String(required=True, load_from='@to')


class OozieForkStartSchema(OozieBaseSchema):
    start = ma.fields.String(required=True, load_from='@start')


class OozieForkSchema(OozieNamedObjectSchema):
    path = ma.fields.List(ma.fields.Nested(OozieForkStartSchema), missing=[])

    singletons_to_lists = ['path']

    @ma.post_load
    def add_base_operator(self, data):
        operator = {
            'name': data['name'],
            'type': 'flow_control',
        }

        copy = data.copy()
        copy['operator'] = operator
        return copy


class OozieJoinSchema(OozieNamedObjectSchema, OozieFlowControlSchema):
    @ma.post_load
    def add_base_operator(self, data):
        operator = {
            'name': data['name'],
            'type': 'flow_control',
        }

        copy = data.copy()
        copy['operator'] = operator
        return copy


class OozieActionSchema(OozieNamedObjectSchema):
    ok = ma.fields.Nested(OozieFlowControlSchema, required=True)
    error = ma.fields.Nested(OozieFlowControlSchema, required=True)

    def _get_action_builder(self, data):
        keyed_action_builders = {
            builder.key: builder
            for builder in self.context['oozie_plugin'].action_builders()
        }

        keys_present = [key for key in keyed_action_builders if data.get(key)]

        if not keys_present:
            raise ma.ValidationError(
                'No known action builders present (expected one of {{{}}}, found keys: {}'.format(
                    ', '.join(keyed_action_builders.keys()),
                    ', '.join(data.keys())))

        if len(keys_present) > 1:
            raise ma.ValidationError(
                'Only one action type may be specified (found {})'.format(
                    keys_present))

        return keyed_action_builders[keys_present[0]]

    @ma.validates_schema(pass_original=True)
    def one_action_type(self, _, original):
        """ Runs the validation checks to make sure that exactly one
            known action is present, and prints an error message based on
            the content of the original, unparsed input
        """
        self._get_action_builder(original)

    @ma.post_load(pass_original=True)
    def fetch_base_action(self, data, original):
        builder_cls = self._get_action_builder(original)

        return builder_cls(self.context, data, original[builder_cls.key])


class OozieKillSchema(OozieNamedObjectSchema):
    message = ma.fields.String()


class OozieCaseSchema(OozieFlowControlSchema):
    text = ma.fields.String(required=True, load_from='#text')


class OozieSwitchSchema(OozieBaseSchema):
    case = ma.fields.List(ma.fields.Nested(OozieCaseSchema))
    default = ma.fields.Nested(OozieFlowControlSchema, required=True)

    singletons_to_lists = ['case']


class OozieDecisionSchema(OozieNamedObjectSchema):
    switch = ma.fields.Nested(OozieSwitchSchema)


class OozieWorkflowAppSchema(OozieNamedObjectSchema):
    name = ma.fields.String(required=True, load_from='@name')
    action = ma.fields.List(ma.fields.Nested(OozieActionSchema), missing=[])
    join = ma.fields.List(ma.fields.Nested(OozieJoinSchema), missing=[])
    fork = ma.fields.List(ma.fields.Nested(OozieForkSchema), missing=[])
    decision = ma.fields.List(ma.fields.Nested(
        OozieDecisionSchema), missing=[])
    start = ma.fields.Nested(OozieFlowControlSchema, required=True)
    end = ma.fields.Nested(OozieNamedObjectSchema, required=True)
    kill = ma.fields.Nested(OozieKillSchema)

    singletons_to_lists = ['fork', 'join', 'action', 'decision']


class OozieWorkflowSchema(OozieBaseSchema):
    workflow_app = ma.fields.Nested(
        OozieWorkflowAppSchema, load_from='workflow-app', required=True)

    @ma.post_load
    def return_workflow(self, data):
        return data['workflow_app']



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import six
import marshmallow as ma
from boundary_layer.schemas.internal.base import BaseSpecSchema
from boundary_layer.schemas.base import StrictSchema


class PropertyPreprocessorSchema(StrictSchema):
    type = ma.fields.String(required=True)
    properties = ma.fields.Dict()
    apply_to_properties = ma.fields.List(ma.fields.String, required=True)


class OperatorSpecSchema(BaseSpecSchema):
    operator_class = ma.fields.String(required=True)
    operator_class_module = ma.fields.String(required=True)

    property_preprocessors = ma.fields.List(ma.fields.Nested(PropertyPreprocessorSchema))

    @ma.validates_schema
    def valid_preprocessor_property_names(self, data):
        preprocessors = data.get('property_preprocessors', [])

        if not preprocessors:
            return

        properties = frozenset(data.get('parameters_jsonschema', {}).get('properties', []))

        for preprocessor in preprocessors:
            missing = [
                property_name for property_name in preprocessor['apply_to_properties']
                if property_name not in properties]

            if missing:
                raise ma.ValidationError(
                    'Properties specified by preprocessor `{}` are not present '
                    'in the schema: `{}`'.format(
                        preprocessor['type'],
                        '`, `'.join(missing)))

        assigned_preprocessors = {
            property_name: [
                preprocessor['type'] for preprocessor in preprocessors
                if property_name in preprocessor['apply_to_properties']]
            for property_name in properties
        }

        dupes = [
            key for (key, value) in six.iteritems(assigned_preprocessors)
            if len(value) > 1]

        if dupes:
            raise ma.ValidationError(
                'One or more properties were assigned multiple preprocessors. '
                'This is not permitted.  Found: {}'.format({
                    key: value for (key, value) in six.iteritems(assigned_preprocessors)
                    if len(value) > 1
                }),
                ['property_preprocessors'])



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import datetime
import boundary_layer
from boundary_layer.builders.base import DagBuilderBase
from boundary_layer.schemas.dag import DagArgsSchema


class PrimaryDagBuilder(DagBuilderBase):
    indent_operators = False

    def preamble(self):
        assert len(self.reference_path) == 1, \
            'Invalid Primary DAG reference path: {}'.format(
                self.reference_path)

        template = self.get_jinja_template('primary_preamble.j2')

        dag_args_dumped = DagArgsSchema(context={'for_dag_output': True}).dump(
            self.dag.get('dag_args', {}))
        if dag_args_dumped.errors:
            # should not happen because the schema was validated upon load,
            # but we should check
            raise Exception('Error serializing dag_args: {}'.format(
                dag_args_dumped.errors))

        dag_args = dag_args_dumped.data
        dag_args['dag_id'] = self.build_dag_id()

        default_task_args = self.dag.get('default_task_args', {})

        return template.render(
            dag_args=dag_args,
            default_task_args=default_task_args,
            imports=self.get_imports(),
            specs=self.specs,
            metadata=self.metadata,
            config_md5=boundary_layer.util.md5(self.metadata.yaml_config),
            build_time=datetime.datetime.utcnow().isoformat(),
            library_name='boundary-layer',
            library_version=boundary_layer.get_version_string(),
        )

    def epilogue(self):
        pass



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

""" Define some exception classes to aid in testing, i.e to make sure that
    workflows are being rejected for appropriate reasons
"""


class CyclicWorkflowException(Exception):
    pass


class DisconnectedWorkflowException(Exception):
    pass


class DuplicateWorkflowNames(Exception):
    pass


class MissingSubWorkflowReference(Exception):
    pass


class NoPrimaryWorkflows(Exception):
    pass


class MultiplePrimaryWorkflows(Exception):
    pass


class InvalidWorkflowReference(Exception):
    pass


class InvalidNodeReference(Exception):
    pass


class InvalidFlowControlNode(Exception):
    pass


class MissingPreprocessorException(Exception):
    pass


class DuplicateRegistryConfigName(Exception):
    pass


class InvalidConfig(Exception):
    pass



import abc


class HadoopClusterConfig(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def managed_resource(self):
        pass

    @abc.abstractproperty
    def apply_config_properties(self, operator_properties, config_properties):
        pass

    @abc.abstractproperty
    def mapreduce_operator_type(self):
        pass

    @abc.abstractproperty
    def hive_operator_type(self):
        pass


class StandardHadoopClusterConfig(HadoopClusterConfig):
    managed_resource = None
    mapreduce_operator_type = 'mapreduce'
    hive_operator_type = 'hive'

    def apply_config_properties(self, operator_properties, config_properties):
        raise NotImplementedError(
            'Sorry, the class `{}` is not fully implemented right now'.format(
                self.__class__.__name__))

    def __init__(self, namenode):
        self.namenode = namenode


class DataprocHadoopClusterConfig(HadoopClusterConfig):
    mapreduce_operator_type = 'dataproc_hadoop'
    hive_operator_type = 'dataproc_hive'

    resource_type = 'dataproc_cluster'
    resource_name = 'dataproc-cluster'

    @property
    def managed_resource(self):
        return {
            'name': self.resource_name,
            'type': self.resource_type,
            'properties': self.create_operator_args,
        }

    def apply_config_properties(self, operator_properties, config_properties):
        result = operator_properties.copy()

        if config_properties:
            result['dataproc_hadoop_properties'] = config_properties

        return result

    def __init__(self, **create_operator_args):
        self.create_operator_args = create_operator_args



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import abc
import marshmallow as ma
from boundary_layer.oozier.schema import OozieBaseSchema


class OozieActionBuilder(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, context, action_metadata, data):
        self.context = context
        self.action_metadata = action_metadata
        self.data = data

    @property
    def ok_to(self):
        return self.action_metadata['ok']['to']

    @property
    def name(self):
        return self.action_metadata['name']

    @abc.abstractproperty
    def key(self):
        pass

    @abc.abstractmethod
    def get_action(self):
        pass

    @abc.abstractmethod
    def get_operator(self):
        pass


class OozieActionBuilderWithSchema(OozieActionBuilder):
    @abc.abstractproperty
    def schema(self):
        pass

    def __init__(self, context, action_metadata, data):
        loaded = self.schema(context=context).load(data)
        if loaded.errors:
            raise ma.ValidationError(loaded.errors)

        super(OozieActionBuilderWithSchema, self).__init__(context, action_metadata, loaded.data)

    def get_action(self):
        result = self.action_metadata.copy()
        result.update(self.data)

        return result


class OozieSubWorkflowActionSchema(OozieBaseSchema):
    app_path = ma.fields.String(required=True, load_from='app-path')
    propagate_configuration = ma.fields.Dict(allow_none=True)


class OozieSubWorkflowBuilderBase(OozieActionBuilderWithSchema):
    schema = OozieSubWorkflowActionSchema

    def __init__(self, context, action_metadata, data):
        super(OozieSubWorkflowBuilderBase, self).__init__(
            context, action_metadata, data)

        # We generally do not know the name of the target sub-workflow
        # a priori; we must load it to find out.  We store the name in this
        # class, and insert it inside the parser using the setter method,
        # below.
        self._target_sub_workflow_name = None

    def set_target_name(self, name):
        self._target_sub_workflow_name = name

    @property
    def target_name(self):
        if not self._target_sub_workflow_name:
            raise Exception(
                'Cannot get sub_workflow_name for action {}: not yet set!'.format(
                    self.get_action()))

        return self._target_sub_workflow_name

    @abc.abstractproperty
    def key(self):
        pass

    def get_operator(self):
        return {
            'name': self.name,
            'type': 'sub_dag',
            'target': self.target_name,
        }



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from copy import deepcopy
import six
from jsonschema import Draft4Validator, validators

# Copy the extend_with_defaults method from the JSON Schema docs:
# http://python-jsonschema.readthedocs.io/en/latest/faq/


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for _property, subschema in six.iteritems(properties):
            if "default" in subschema:
                instance.setdefault(_property, subschema["default"])

        for error in validate_properties(
                validator, properties, instance, schema):
            yield error

    return validators.extend(
        validator_class, {"properties": set_defaults},
    )


DefaultValidatingDraft4Validator = extend_with_default(Draft4Validator)


def validate_and_fill_defaults(item, schema):
    copy = deepcopy(item)

    DefaultValidatingDraft4Validator(schema).validate(copy)

    return copy



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from collections import namedtuple
from boundary_layer.oozier.cluster_config import DataprocHadoopClusterConfig
from boundary_layer.plugins import BaseOozieParserPlugin
from .oozie_actions import \
        OozieSubWorkflowBuilder, \
        OozieFileSystemActionBuilder, \
        OozieSshActionBuilder, \
        OozieMapReduceActionBuilder


ExternalTaskSpec = namedtuple('ExternalTaskSpec', ['dag_id', 'task_id'])


def _external_task_spec(item):
    """ A parser for the --with-external-task-sensors argument """
    if item.count(':') != 1:
        raise ValueError(
            'Invalid external task specification `{}`: expected a '
            'string of the form <dag_id>:<task_id>'.format(item))

    return ExternalTaskSpec(*item.split(':'))


class DefaultOozieParserPlugin(BaseOozieParserPlugin):
    @classmethod
    def register_arguments(cls, parser):
        parser.add_argument('--with-external-task-sensors',
                            type=_external_task_spec,
                            nargs='+',
                            default=[])

        parser.add_argument(
            '--dag-max-active-runs',
            type=int,
            help='Argument for DAG max_active_runs parameter')
        parser.add_argument(
            '--dag-schedule-interval',
            help='Argument for DAG schedule_interval parameter')
        parser.add_argument(
            '--dag-disable-catchup',
            default=False,
            action='store_true',
            help='Whether to set the DAG parameter catchup=False')
        parser.add_argument(
            '--dag-concurrency',
            type=int,
            help='argument for DAG concurrency parameter')

        dataproc_group = parser.add_argument_group('Arguments for dataproc cluster configuration')
        dataproc_group.add_argument('--cluster-depends-on-past', default=None, action='store_true')
        dataproc_group.add_argument(
            '--cluster-wait-for-downstream',
            default=None,
            action='store_true')

        dataproc_group.add_argument('--cluster-num-workers', type=int, default=128)
        dataproc_group.add_argument('--cluster-base-name', default=None)
        dataproc_group.add_argument('--cluster-name-suffix', default=None)
        dataproc_group.add_argument(
            '--cluster-project-id',
            default=None,
            help='The GCP project in which to run the dataproc cluster')
        dataproc_group.add_argument(
            '--cluster-region',
            default=None,
            help='The GCP region in which to run the dataproc cluster')

    def action_builders(self):
        return [
            OozieSubWorkflowBuilder,
            OozieFileSystemActionBuilder,
            OozieSshActionBuilder,
            OozieMapReduceActionBuilder,
        ]

    def upstream_operators(self):
        result = []
        for sensor in self.args.with_external_task_sensors:
            result.append({
                'name': '{}-{}-sensor'.format(sensor.dag_id, sensor.task_id),
                'type': 'external_task_sensor',
                'properties': {
                    'external_dag_id': sensor.dag_id,
                    'external_task_id': sensor.task_id,
                }
            })

        return result

    def jsp_macros(self):
        return {
            'wf:id()': '{{ task.task_id }}',
            'wf:name()': '{{ task.task_id }}',
        }

    def dag_args(self):
        result = {}

        if self.args.dag_max_active_runs:
            result['max_active_runs'] = self.args.dag_max_active_runs

        if self.args.dag_schedule_interval:
            result['schedule_interval'] = self.args.dag_schedule_interval

        result['catchup'] = not self.args.dag_disable_catchup

        if self.args.dag_concurrency:
            result['concurrency'] = self.args.dag_concurrency

        return result

    def cluster_config(self):
        cluster_base_name = self.args.cluster_base_name or \
            self.args.workflow_name

        cluster_name_suffix = self.args.cluster_name_suffix or "{{ ds }}"

        cluster_properties = {
            'cluster_name': cluster_base_name + "-" + cluster_name_suffix,
            'num_workers': self.args.cluster_num_workers,
        }

        if self.args.cluster_project_id:
            cluster_properties['project_id'] = self.args.cluster_project_id

        if self.args.cluster_region:
            cluster_properties['region'] = self.args.cluster_region

        if self.args.cluster_depends_on_past:
            cluster_properties['depends_on_past'] = True

        if self.args.cluster_wait_for_downstream:
            cluster_properties['wait_for_downstream'] = True

        return DataprocHadoopClusterConfig(**cluster_properties)



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import pkg_resources
import semver


def get_version_string():
    version = pkg_resources.get_distribution('boundary_layer').version

    # Pip seems to replace '-' with '.' in the version strings, for some reason.
    # this makes semver unhappy, so we must replace .dev0 with -dev0
    if not version.endswith('.dev0'):
        return version

    dev0_position = version.rindex('.dev0')
    return version[:dev0_position] + '-' + version[1 + dev0_position:]


VERSION_STRING = get_version_string()
VERSION = semver.parse_version_info(VERSION_STRING)
MIN_SUPPORTED_VERSION = semver.parse_version_info('0.9.9')

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import sys
import logging
from boundary_layer import VERSION_STRING


def setup_logger():
    result = logging.getLogger('boundary-layer v. {}'.format(VERSION_STRING))

    stream = logging.StreamHandler(sys.stderr)

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    stream.setFormatter(formatter)

    result.addHandler(stream)
    result.setLevel(logging.INFO)

    return result


logger = setup_logger()



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import abc
from six.moves import filter
from jinja2 import Environment, PackageLoader
from boundary_layer.builders import util
from boundary_layer.logger import logger
from boundary_layer.registry import NodeTypes
from boundary_layer.util import sanitize_operator_name
from boundary_layer.containers import WorkflowMetadata


class DagBuilderBase(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def indent_operators(self):
        pass

    @abc.abstractmethod
    def preamble(self):
        pass

    @abc.abstractmethod
    def epilogue(self):
        pass

    @staticmethod
    def _build_jinja_env():
        jenv = Environment(
            loader=PackageLoader('boundary_layer', 'builders/templates'),
            trim_blocks=True)

        jenv.filters['format_value'] = util.format_value
        jenv.filters['add_leading_spaces'] = util.add_leading_spaces
        jenv.filters['comment'] = util.comment
        jenv.filters['sanitize_operator_name'] = sanitize_operator_name
        jenv.filters['verbatim'] = util.verbatim

        return jenv

    def get_jinja_template(self, template_filename):
        return self._build_jinja_env().get_template(template_filename)

    def get_imports(self):
        all_nodes = self.specs.graphs.primary.ordered() + \
            [node for graph in self.specs.graphs.secondary
             for node in graph.ordered()]

        all_imports = [self.dag.get('imports', {})] + \
            [node.imports() for node in all_nodes]

        objects = {}
        modules = set()
        for node_imports in all_imports:
            modules |= set(node_imports.get('modules', []))

            for item in node_imports.get('objects', []):
                objects.setdefault(item['module'], set())
                objects[item['module']] |= set(item['objects'])

        return {
            'modules': modules,
            'objects': objects,
        }

    def __init__(
            self,
            dag,
            graph,
            reference_path,
            specs,
            metadata=None,
            referring_node=None,
            sub_dag_builder=None,
            generator_builder=None):
        self.dag = dag
        self.graph = graph
        self.reference_path = reference_path
        self.specs = specs
        self.metadata = metadata or WorkflowMetadata(None, None)
        self.referring_node = referring_node

        self.sub_dag_builder = sub_dag_builder
        self.generator_builder = generator_builder

    @property
    def default_task_args(self):
        return self.specs.parsed.primary.get('default_task_args', {})

    def build_dag_id(self):
        return util.construct_dag_name(self.reference_path)

    def render_operator(self, node):
        template_filename = None
        if node.type == NodeTypes.GENERATOR:
            template_filename = 'generator_operator.j2'
        elif node.type == NodeTypes.SUBDAG:
            template_filename = 'subdag_operator.j2'
        else:
            template_filename = 'operator.j2'

        template = self.get_jinja_template(template_filename)

        # Do not set upstream/downstream dependencies that involve generator nodes
        # at this stage; those are all set within the generator nodes, and if they are
        # set here, there will be python errors due to references to operators that
        # do not exist (generators do not correspond to operators)
        generator_nodes = frozenset(
            gen.name for gen in self.graph.graph.nodes
            if gen.type == NodeTypes.GENERATOR)

        upstream_deps = frozenset(
            dep.name for dep in self.graph.upstream_dependency_set(node))

        if generator_nodes & upstream_deps:
            logger.debug(
                'Not passing upstream generator dependencies `%s` to '
                'operator template for node `%s`',
                generator_nodes & upstream_deps,
                node.name)

        downstream_deps = frozenset(
            dep.name for dep in self.graph.downstream_dependency_set(node))

        if generator_nodes & downstream_deps:
            logger.debug(
                'Not passing downstream generator dependencies `%s` to '
                'operator template for node `%s`',
                generator_nodes & downstream_deps,
                node.name)

        return template.render(
            node=node,
            args=node.operator_args,
            upstream_dependencies=list(upstream_deps - generator_nodes),
            downstream_dependencies=list(downstream_deps - generator_nodes),
        )

    def get_secondary_dag(self, target):
        hits = [dag for dag in self.specs.parsed.secondary
                if dag['name'] == target]

        if not hits:
            raise ValueError('Secondary dag id {} not found'.format(target))

        if len(hits) > 1:
            raise ValueError(
                'Multiple hits for secondary dag id {}'.format(target))

        return hits[0]

    def get_secondary_graph(self, target):
        """ Get the graph corresponding to the target.  This is kind of ugly,
            a consequence of the way in which we currently store dags separately
            from graphs.  Ideally there would be only one of the two methods,
            get_secondary_(dag|graph).
        """
        self.get_secondary_dag(target)  # does the checking

        for (idx, dag) in enumerate(self.specs.parsed.secondary):
            if dag['name'] == target:
                return self.specs.graphs.secondary[idx]

        raise Exception("should not be possible")

    def get_target_builder_cls(self, node_type):
        if node_type == NodeTypes.GENERATOR:
            if not self.generator_builder:
                raise Exception('No generator builder is defined!')

            return self.generator_builder

        elif node_type == NodeTypes.SUBDAG:
            if not self.sub_dag_builder:
                raise Exception('No sub_dag builder is defined!')

            return self.sub_dag_builder

        raise Exception(
            'Node type `{}` has no known target builder'.format(
                node_type))

    def render_target(self, node):
        builder = self.get_target_builder_cls(node.type)(
            dag=self.get_secondary_dag(node.target),
            graph=self.get_secondary_graph(node.target),
            reference_path=self.reference_path + [node.name],
            specs=self.specs,
            referring_node=node,
            sub_dag_builder=self.sub_dag_builder,
            generator_builder=self.generator_builder,
            )

        return builder.build()

    def build(self):
        # Keep track of which subdag and generator targets have been rendered.
        # These targets can be reused by multiple referring nodes.
        rendered_targets = set()

        # We build the result by appending components to an array and then
        # joining together at the end
        components = [self.preamble()]

        # generators are rendered last, because they refer to both upstream and
        # downstream components when they express their dependencies
        generator_components = []

        for node in self.graph.ordered():
            operator = None
            if node.type in set([NodeTypes.GENERATOR, NodeTypes.SUBDAG]) \
                    and node.target not in rendered_targets:

                operator = '\n'.join([
                    self.render_target(node),
                    self.render_operator(node)])

                rendered_targets.add(node.target)

            elif node.type in NodeTypes:
                operator = self.render_operator(node)

            else:
                raise Exception(
                    'Unrecognized operator type: {}'.format(node.type))

            # add the rendered operator to the appropriate components list
            (components if node.type != NodeTypes.GENERATOR else generator_components).append(
                util.add_leading_spaces(
                    operator,
                    1 if self.indent_operators else 0))

        components += generator_components
        components.append(self.epilogue())

        return '\n'.join(filter(None, components))



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re
import six
from six.moves import map
import marshmallow as ma


class JspMacroTranslator(object):
    # Oozie macros use JSP Expression Language.  I can't find a convenient
    # expression evaluator, but there aren't too many things we need to support
    # so we can do it with regexes and substitutions
    OOZIE_MACRO_RE = re.compile(r'\$\{[^}]+\}')

    def __init__(self, macros):
        self.macros = macros

    def translate(self, item):
        if isinstance(item, dict):
            return {
                key: self.translate(value) for (key, value) in six.iteritems(item)
            }
        elif isinstance(item, list):
            return list(map(self.translate, item))

        if not isinstance(item, six.string_types):
            raise ma.ValidationError('Cannot translate object {} of type {}'.format(
                item,
                type(item)))

        hits = list(self.OOZIE_MACRO_RE.finditer(item))
        if not hits:
            return item

        split = self.OOZIE_MACRO_RE.split(item)
        # Some sanity checking to make sure that the split gave us what we expected.
        # Given that we know there's at least one hit, we expect that there should be
        # at least one more item in splits than in hits.  If the text field both
        # begins and ends with a hit, then there can potentially be 2 more items
        # in the split array.
        if not (2 + len(hits) >= len(split) > len(hits)):
            raise ma.ValidationError(
                'Bad split on input text `{}`: hits == `{}` / split == `{}`'.format(
                    item,
                    hits,
                    split))

        if not(split[-1] == '' or len(split) == 1 + len(hits)):
            raise ma.ValidationError(
                'Bad split on input text `{}`: hits == `{}` / split == `{}`'.format(
                    item,
                    hits,
                    split))

        result = split[0]
        for (idx, hit) in enumerate(hits):
            # Must discard the first 2 and final 1 characters from the hit, because
            # these represent the '${' and '}'.  Note that we could use capture groups
            # in the regex but then these will contaminate the output of the split()
            # function (see: https://docs.python.org/2/library/re.html#re.split)
            assert hit.group(0).startswith('${') and hit.group(0).endswith('}')
            oozie_macro = hit.group(0)[2:-1]
            translation = self.macros.get(oozie_macro)

            if translation is None:
                raise ma.ValidationError(
                    'Could not translate {}: unrecognized macro `{}`'.format(item, oozie_macro))

            # if we've made it here, either translation was successful or
            # unsuccessful translations are permitted.  In the latter case,
            # we reinsert the oozie macro.
            result += translation or hit.group(0)
            result += split[1 + idx]

        return result



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import six
from boundary_layer.logger import logger
from boundary_layer.registry import ConfigFileRegistry, RegistryNode, NodeTypes
from boundary_layer.schemas.internal.operators import OperatorSpecSchema
from boundary_layer.schemas.dag import ImportSchema
from boundary_layer import validator, util
from boundary_layer.containers import PropertySources, ResolvedProperties
from boundary_layer.exceptions import MissingPreprocessorException


class OperatorNode(RegistryNode):
    type = NodeTypes.OPERATOR

    @property
    def operator_class(self):
        return self.config['operator_class']

    @property
    def operator_class_module(self):
        return self.config['operator_class_module']

    def __init__(self, config, item):
        super(OperatorNode, self).__init__(config=config, item=item)

        self._resolved_properties = None
        self._preprocessor_imports = None
        self._default_task_args = None

    def set_default_task_args(self, args):
        self._default_task_args = args or {}

    @property
    def resolved_properties(self):
        if not self._resolved_properties:
            raise Exception(
                'Cannot retrieve resolved properties for operator {}: '
                'resolve_properties() has not been called yet!'.format(
                    self))

        return self._resolved_properties

    @property
    def operator_args(self):
        """ Return a dictionary of arguments to apply to the operator constructor.
            This method accounts for the existence of default_task_args, by
            removing any operator args that will be set from the default task args.
            This way, any costly constructions in the default task args are not
            re-computed for each operator that uses them.
        """
        resolved = self.resolved_properties

        result = resolved.values.copy()

        for property_name in resolved.sources.default_task_args:
            if self._default_task_args[property_name] == resolved.values[property_name]:
                result.pop(property_name)

        return result

    def imports(self):
        if not self._resolved_properties:
            raise Exception(
                'Cannot retrieve imports for operator {}: '
                'resolve_properties() has not been called yet!'.format(
                    self))

        loaded = ImportSchema().load(self.config.get('imports', {}))

        assert not loaded.errors, \
            ('Internal error: processing `imports` config {} for '
             'operator `{}`').format(
                 self.config.get('imports', {}),
                 self.name)

        result = loaded.data

        if self.operator_class:
            result.setdefault('objects', [])
            result['objects'].append({
                'module': self.operator_class_module,
                'objects': [self.operator_class],
            })

        for item in self._preprocessor_imports.values():
            if 'objects' in item:
                result.setdefault('objects', [])
                result['objects'] += item['objects']

            if 'modules' in item:
                result.setdefault('modules', [])
                result['modules'] += item['modules']

        return result

    def resolve_properties(
            self,
            execution_context,
            default_task_args=None,
            base_operator_loader=None,
            preprocessor_loader=None):
        """ Get the properties / arguments for the operator, and split them
            according to their source.  Specifically, properties are provided
            to the operator by either the DAG config file, the resources
            available in the operator's context, any task defaults specified
            in the primary DAG, and the schema defaults, in that order of
            precedence.

            Once the properties are all resolved, this method then validates
            all of the resolved arguments against the task's schema.

            :param execution_context: the context in which this node is executed,
                specifically containing the available resources and the node
                that referred to this node, if any
            :type execution_context: boundary_layer.containers.ExecutionContext
            :param default_task_args: the default task args defined in the
                DAG
            :type default_task_args: dict
            :param base_operator_loader: A method that retrieves typed operators,
                equivalent to a Registry.get method
            :type base_operator_loader: callable
            :param preprocessor_loader: A method that retrieves typed preprocessors,
                equivalent to a Registry.get method
            :type preprocessor_loader: callable

            :returns: a mapping of property source to property key/value pairs
            :rtype: dict<dict<string, any>>
        """
        schema = self.get_schema(base_operator_loader)
        schema_properties = frozenset(schema.get('properties', {}).keys())

        self.set_default_task_args(default_task_args)

        (sources, property_values) = self._get_property_sources_and_values(
            schema_properties,
            execution_context)

        validated = validator.validate_and_fill_defaults(
            item=property_values,
            schema=schema)

        for key in validated:
            if key not in property_values:
                continue

            sources.schema.add(key)

        logger.debug('%s: validated partitioned properties: %s', self.name, sources)

        preprocessors = self._load_preprocessors(
            base_operator_loader,
            preprocessor_loader)

        self._preprocessor_imports = {
            pp_name: pp.imports()
            for (pp_name, pp) in six.iteritems(preprocessors)
        }

        preprocessed_values = self._apply_preprocessors(
            args=validated,
            preprocessors=preprocessors)

        if self._resolved_properties:
            if preprocessed_values != self._resolved_properties.values:
                raise Exception(
                    'resolve_properties() was already called for operator {}, '
                    'and different values were computed this time!  Found: {}, '
                    'expected: {}.  This was probably caused by repeated '
                    'references to a sub-dag or generator using different resource '
                    'contexts.  This is not presently supported!'.format(
                        self,
                        preprocessed_values,
                        self._resolved_properties.values))
            else:
                logger.warning(
                    'resolve_properties() was already called for operator %s, '
                    'but no differences in the computed properties were found.',
                    self)

        self._resolved_properties = ResolvedProperties(
            sources=sources,
            values=preprocessed_values)

        return self._resolved_properties

    def _get_property_sources_and_values(
            self,
            schema_properties,
            execution_context):
        """ For the provided set of properties, determine the values, and the
            sources of those values, for the current node.

            Value sources include:
                 - default task args defined in the DAG
                 - available resources
                 - properties defined in the DAG
                 - fixed values (set below)
                 - global defaults (set below)

            Note that this method does not validate these values against the
            schema, and therefore does not include any values that could be
            derived from the schema's default settings.

            :param schema_properties: the list of property names that are
                applicable to this node
            :type schema_properties: list<str>
            :param default_task_args: default arg values provided in the DAG
            :type default_task_args: dict
            :param execution_context: the execution context
            :type execution_context: boundary_layer.containers.ExecutionContext
            :returns: sources and values as a 2-tuple
            :rtype: (PropertySources, dict)
        """

        sources = PropertySources(
            dag=set(),
            default_task_args=set(),
            resources=set(),
            schema=set(),
            global_defaults=set(),
            fixed_args=set(),
            unknown_to_schema=set())

        property_values = {}

        resource_args = self._get_resource_args(execution_context)

        global_defaults = self._get_global_defaults(execution_context)

        fixed_args = self._get_fixed_args()

        # make sure that the user has not tried to specify values for any
        # fixed args; this prevents the user from trying to attach operators
        # to a different DAG, for instance (which does not make sense because
        # there is only one DAG)
        invalid_properties = [
            property_name for property_name in fixed_args
            if property_name in self.properties]

        if invalid_properties:
            raise Exception(
                'Illegal properties `{}` provided for operator `{}`: these '
                'properties are assigned fixed values by boundary-layer that '
                'cannot be overridden'.format(
                    '` ,`'.join(invalid_properties),
                    self))

        for property_name in schema_properties:
            if property_name in fixed_args:
                # Check fixed args first, because we do not allow these to be
                # set by the user
                value = fixed_args[property_name]
                logger.debug(
                    '%s: Inserting value `%s` for argument `%s` from fixed_args',
                    self.name, value, property_name)

                property_values[property_name] = value
                sources.fixed_args.add(property_name)
                continue

            if property_name in self.properties:
                logger.debug(
                    '%s: Property `%s` found in user-props', self.name, property_name)

                property_values[property_name] = self.properties[property_name]
                sources.dag.add(property_name)
                continue

            resource_hits = resource_args.get(property_name, {})

            if len(resource_hits) > 1:
                raise ValueError('Error in operator {}: Multiple available resources '
                                 'provide the argument {}: {}. Please specify a value or limit '
                                 'limit the resource scope'.format(
                                     self.name,
                                     property_name,
                                     resource_hits))

            if len(resource_hits) == 1:
                (resource_name, value) = resource_hits.popitem()

                logger.debug('%s: Inserting value `%s` for argument `%s` from resource `%s`',
                             self.name, value, property_name, resource_name)

                property_values[property_name] = value
                sources.resources.add(property_name)
                continue

            if property_name in self._default_task_args:
                value = self._default_task_args[property_name]
                logger.debug('%s: Inserting value `%s` for argument `%s` from default_task_args',
                             self.name, value, property_name)

                property_values[property_name] = value
                sources.default_task_args.add(property_name)
                continue

            if property_name in global_defaults:
                value = global_defaults[property_name]
                logger.debug('%s: Inserting value `%s` for argument `%s` from global defaults',
                             self.name, value, property_name)
                property_values[property_name] = value
                sources.global_defaults.add(property_name)
                continue

            logger.debug(
                '%s: No resources or defaults available for property `%s`',
                self.name,
                property_name)

        unknown_to_schema = [
            property_name for property_name in self.properties
            if property_name not in schema_properties]

        for property_name in unknown_to_schema:
            value = self.properties[property_name]
            logger.debug(
                '%s: Inserting value `%s` for user-property `%s` which is not '
                'part of the schema for this operator',
                self.name,
                value,
                property_name)
            property_values[property_name] = value
            sources.unknown_to_schema.add(property_name)

        return (sources, property_values)

    def _apply_preprocessors(self, args, preprocessors):
        """ Apply any necessary preprocessing to the alread-validated args.
            This must be the last step in case any preprocessors are defined on
            fields that are inserted by the schema defaults.
        """
        result = args.copy()

        for (property_name, preprocessor) in six.iteritems(preprocessors):
            if property_name not in args:
                continue

            processed_value = preprocessor.process_arg(
                args[property_name],
                node=self,
                raw_args=args)
            logger.debug(
                'Property `%s` raw value: `%s`, processed value: `%s`',
                property_name,
                args[property_name],
                processed_value)

            result[property_name] = processed_value

        return result

    def _get_resource_args(self, execution_context):
        resources_available = self._get_resources_available(execution_context)
        result = {}

        for (resource_name, resource) in six.iteritems(resources_available):
            for (property_name, value) in six.iteritems(resource.get_provided_args()):
                result.setdefault(property_name, {})
                result[property_name][resource_name] = value

        return result

    def _get_resources_available(self, execution_context):
        keys_available = self.requires_resources & frozenset(execution_context.resources)
        return {key: execution_context.resources[key] for key in keys_available}

    def _load_preprocessors(self, base_loader, preprocessor_loader):
        def aggregator(previous_result, node):
            return previous_result + node.config.get('property_preprocessors', [])
        preprocessor_configs = self._aggregate_over_hierarchy(
            base_loader=base_loader,
            initial_value=self.config.get('property_preprocessors', []),
            aggregator=aggregator)

        if not preprocessor_configs:
            return {}

        if not preprocessor_loader:
            raise MissingPreprocessorException(
                'Node {} of type {} requires preprocessors {}, but no '
                'preprocessor_loader is available!'.format(
                    self,
                    self.type,
                    [config['type'] for config in preprocessor_configs]))

        result = {}
        for preprocessor_conf in preprocessor_configs:
            preprocessor = preprocessor_loader(preprocessor_conf)

            for property_name in preprocessor_conf['apply_to_properties']:
                result[property_name] = preprocessor
        return result

    def _get_fixed_args(self):
        return {
            'dag': '<<dag>>',
        }

    def _get_global_defaults(self, execution_context):
        return {
            'task_id': self._build_task_id(execution_context),
        }

    def _build_task_id(self, execution_context):
        base_name = util.sanitize_operator_name(self.name)
        if not execution_context.referrer or execution_context.referrer.type != NodeTypes.GENERATOR:
            return base_name

        suffix_mode = execution_context.referrer.item.get('auto_task_id_mode')
        if not suffix_mode or suffix_mode == 'item_name':
            return base_name + '-<<item_name>>'
        elif suffix_mode == 'index':
            return base_name + '-<<str(index)>>'

        raise Exception(
            'Unknown suffix_mode `{}` for generator `{}` found while processing '
            'node `{}`'.format(
                suffix_mode,
                execution_context.referrer.name,
                self.name))


class OperatorRegistry(ConfigFileRegistry):
    node_cls = OperatorNode
    spec_schema_cls = OperatorSpecSchema



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from jsonschema import Draft4Validator
import marshmallow as ma
from boundary_layer.logger import logger
from boundary_layer.schemas.base import StrictSchema
from boundary_layer.schemas.dag import ImportSchema


class JsonSchemaSchema(StrictSchema):
    properties = ma.fields.Dict()
    additionalProperties = ma.fields.Boolean()
    required = ma.fields.List(ma.fields.String())


class BaseSpecSchema(StrictSchema):
    name = ma.fields.String(required=True)
    imports = ma.fields.Nested(ImportSchema)
    parameters_jsonschema = ma.fields.Nested(JsonSchemaSchema)
    schema_extends = ma.fields.String()

    @ma.validates_schema
    def jsonschema_or_extended(self, data):
        if 'parameters_jsonschema' in data or 'schema_extends' in data:
            return

        raise ma.ValidationError(
            'No mechanism for determining properties schema for `{}`: Must provide '
            'at least one of (parameters_jsonschema, schema_extends)'.format(
                data.get('name', data)))

    @ma.validates_schema
    def check_jsonschema(self, data):
        if 'parameters_jsonschema' not in data:
            return
        # Make sure that `properties` is present, because it's not actually
        # required to make a valid JSONSchema
        if 'properties' not in data.get('parameters_jsonschema', {}):
            logger.debug(
                'No `properties` defined in `parameters_jsonschema` for `%s`',
                data['name'])
        try:
            Draft4Validator.check_schema(data['parameters_jsonschema'])
        except Exception as e:
            raise ma.ValidationError('Invalid JSON schema: {}'.format(e),
                                     ['parameters_jsonschema'])



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import itertools
import six


def merge_dicts(dicts):
    result = {}

    for _dict in dicts:
        bad_keys = [key for (key, value) in six.iteritems(_dict) if
                    result.get(key, value) != value]
        if bad_keys:
            raise Exception('ERROR: plugins define incompatible values for keys {}'.format(
                bad_keys))

        result.update(_dict)

    return result


def merge_lists(lists):
    return list(itertools.chain.from_iterable(lists))



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import marshmallow as ma
from boundary_layer.oozier.schema import OozieBaseSchema
from boundary_layer.oozier.actions import OozieActionBuilderWithSchema, OozieSubWorkflowBuilderBase


class OozieSubWorkflowBuilder(OozieSubWorkflowBuilderBase):
    key = 'sub-workflow'


class OozieFileSystemSourceDestSchema(OozieBaseSchema):
    source = ma.fields.String(required=True)
    target = ma.fields.String(required=True)


class OozieFileSystemPathSchema(OozieBaseSchema):
    path = ma.fields.String(required=True, load_from='@path')


class OozieFileSystemActionSchema(OozieBaseSchema):
    delete = ma.fields.List(ma.fields.Nested(OozieFileSystemPathSchema))
    mkdir = ma.fields.List(ma.fields.Nested(OozieFileSystemPathSchema))
    move = ma.fields.List(ma.fields.Nested(OozieFileSystemSourceDestSchema))

    singletons_to_lists = ['delete', 'mkdir', 'move']


class OozieFileSystemActionBuilder(OozieActionBuilderWithSchema):
    key = 'fs'
    schema = OozieFileSystemActionSchema

    def get_operator(self):
        return {
            'name': self.name,
            'type': 'file_system',
            'properties': {}
        }


class OozieSshActionSchema(OozieBaseSchema):
    args = ma.fields.List(ma.fields.String(), missing=[])
    capture_output = ma.fields.Boolean(allow_none=True)
    command = ma.fields.String(required=True)
    host = ma.fields.String(required=True)


class OozieSshActionBuilder(OozieActionBuilderWithSchema):
    key = 'ssh'
    schema = OozieSshActionSchema

    def get_operator(self):
        properties = {
            'remote_host': self.data['host'],
            'command': ' '.join([self.data['command']] + self.data['args']),
        }
        return {
            'name': self.name,
            'type': 'ssh',
            'properties': properties,
        }


class OozieNameValueSchema(OozieBaseSchema):
    name = ma.fields.String(required=True)
    value = ma.fields.String(required=True)


class OozieHadoopConfigurationSchema(OozieBaseSchema):
    _property = ma.fields.List(
        ma.fields.Nested(OozieNameValueSchema),
        required=True,
        load_from='property',
        dump_to='property',
        attribute='property')

    singletons_to_lists = ['property']


class OozieMapReduceActionSchema(OozieBaseSchema):
    arg = ma.fields.List(ma.fields.String(), missing=[])
    configuration = ma.fields.Nested(OozieHadoopConfigurationSchema)
    job_tracker = ma.fields.String(required=True, load_from='job-tracker')
    name_node = ma.fields.String(required=True, load_from='name-node')
    main_class = ma.fields.String(load_from='main-class')


class OozieMapReduceActionBuilder(OozieActionBuilderWithSchema):
    key = 'map-reduce'
    schema = OozieMapReduceActionSchema

    discard_property_keys = []

    def translated_args(self):
        return self.context['macro_translator'].translate(self.data['arg'])

    def operator_properties(self):
        properties = {'arguments': self.translated_args()}

        if self.data.get('main_class'):
            properties['main_class'] = self.data['main_class']

        config_properties = {}
        if self.data.get('configuration', {}).get('property'):
            config_properties = self.context['macro_translator'].translate({
                item['name']: item['value']
                for item in self.data['configuration']['property']
                if item['name'] not in self.discard_property_keys
            })

            if config_properties:
                properties['dataproc_hadoop_properties'] = config_properties

        cluster_config = self.context['cluster_config']
        return cluster_config.apply_config_properties(properties, config_properties)

    def get_operator(self):
        cluster_config = self.context['cluster_config']
        operator = {
            'name': self.name,
            'type': cluster_config.mapreduce_operator_type,
            'properties': self.operator_properties(),
        }

        if cluster_config.managed_resource:
            operator['requires_resources'] = [
                cluster_config.managed_resource['name']
            ]

        return operator



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import abc
import os
from enum import Enum
import yaml

from boundary_layer.logger import logger
from boundary_layer import util
from boundary_layer.exceptions import DuplicateRegistryConfigName, InvalidConfig


class NodeTypes(Enum):
    OPERATOR = 'operator'
    RESOURCE = 'resource'
    GENERATOR = 'generator'
    SUBDAG = 'sub_dag'
    PREPROCESSOR = 'preprocessor'


class RegistryNode(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def type(self):
        pass

    @abc.abstractmethod
    def imports(self):
        pass

    def __init__(self, config, item):
        self.config = config
        self.item = item

        if self.type not in NodeTypes:
            raise ValueError('Invalid package type `{}`'.format(self.type))

        self._icebox = None

    @property
    def name(self):
        return self.item['name']

    @property
    def upstream_dependencies(self):
        return frozenset(self.item.get('upstream_dependencies', []))

    @property
    def downstream_dependencies(self):
        return frozenset(self.item.get('downstream_dependencies', []))

    @property
    def requires_resources(self):
        return frozenset(self.item.get('requires_resources', []))

    @property
    def properties(self):
        return self.item.get('properties', {})

    def _aggregate_over_hierarchy(self, base_loader, initial_value, aggregator):
        stack = [self]
        parent_set = set()
        result = initial_value
        while stack:
            node = stack.pop()
            schema_extends = node.config.get('schema_extends')
            if not schema_extends:
                continue

            if schema_extends in parent_set:
                raise Exception(
                    'Circular reference in schema_extends for node {}: '
                    'repeated parent {}'.format(self, schema_extends))

            parent_set.add(schema_extends)

            parent_node = base_loader({'type': schema_extends})
            stack.append(parent_node)

            result = aggregator(result, parent_node)

        return result

    def get_schema(self, base_loader):
        def aggregator(previous_result, node):
            schema = node.config.get('parameters_jsonschema', {})

            return util.merge_schemas(parent=schema, child=previous_result)

        return self._aggregate_over_hierarchy(
            base_loader=base_loader,
            initial_value=self.config.get('parameters_jsonschema', {}),
            aggregator=aggregator)

    def freeze(self):
        if self._icebox:
            return self._icebox

        self._icebox = (util.freeze(self.config), util.freeze(self.item))
        return self._icebox

    def __hash__(self):
        return hash(self.freeze())

    def __eq__(self, that):
        return isinstance(that, self.__class__) and \
            self.freeze() == that.freeze()

    def __ne__(self, that):
        return not self.__eq__(that)

    def __repr__(self):
        return '{self.__class__.__name__}({self.name})'.format(self=self)


class Registry(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def node_cls(self):
        pass

    @abc.abstractproperty
    def node_configs(self):
        pass

    def get(self, item):
        if 'type' not in item:
            raise Exception(
                'Invalid item: no `type` specified for `{}`'.format(item))

        if item['type'] not in self.node_configs:
            logger.debug('Unrecognized type `%s` for item %s',
                         item['type'], item)
            return None

        return self.node_cls(config=self.node_configs[item['type']], item=item)


class ConfigFileRegistry(Registry):
    @abc.abstractproperty
    def spec_schema_cls(self):
        pass

    def __init__(self, config_paths):
        self.config_paths = config_paths

        self._node_configs = self.load_configs(config_paths)

        if not issubclass(self.node_cls, RegistryNode):
            raise Exception('Invalid `node_cls` property {} (must be a '
                            'subclass of RegistryNode)'.format(
                                self.node_cls))

    @property
    def node_configs(self):
        return self._node_configs

    def load_from_file(self, filename):
        item = None
        with open(filename) as _in:
            item = yaml.safe_load(_in)

        logger.debug('validating item %s against schema %s',
                     item, self.spec_schema_cls.__name__)

        loaded = self.spec_schema_cls().load(item)
        if loaded.errors:
            raise InvalidConfig('Invalid config spec in file {}: {}'.format(
                filename, loaded.errors))

        return loaded.data

    def load_configs(self, config_paths):
        if not isinstance(config_paths, list):
            raise TypeError(
                'Invalid `config_paths` argument: expected a '
                'list<str>, got {}'.format(config_paths))

        registry = {}
        duplicates = {}

        for path in config_paths:
            logger.debug('Loading configs from path %s', path)
            for filename in os.listdir(path):
                full_path = os.path.abspath(os.path.join(path, filename))

                if os.path.splitext(filename)[1] not in ['.yaml', '.yml']:
                    raise Exception(
                        'Invalid file found: {} (must be a yaml config file)'.format(
                            full_path))
                config = self.load_from_file(full_path)

                name = config['name']

                if name in registry:
                    duplicates.setdefault(name, [registry[name]])
                    duplicates[name].append(config)
                    continue

                registry[name] = config

        if duplicates:
            raise DuplicateRegistryConfigName(
                'Duplicate names found while loading registry: `{}` '
                '(full configurations: {})'.format(
                    '`, `'.join(duplicates.keys()),
                    duplicates))

        return registry



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import abc
import base64
import os


class BaseFileFetcher(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def fetch_file_content(self, rel_path):
        pass


class LocalFileFetcher(BaseFileFetcher):
    def __init__(self, base_path):
        self.base_path = base_path

    def fetch_file_content(self, rel_path):
        full_path = os.path.join(self.base_path, rel_path.lstrip('/'))

        content = None
        with open(full_path) as _in:
            content = _in.read()

        return content


class GithubFileFetcher(BaseFileFetcher):
    """ A fetcher for files stored in Github.
        :param base_path: base path of files in the repository
        :type base_path: str
        :param repository: an already-initialized github3 Repository object
        :type repository: github3.repos.repo.Repository
        :param ref: branch from which to fetch files
            Default: master
        :type ref: str
    """

    def __init__(
            self,
            base_path,
            repo_owner,
            repo_name,
            repo_ref=None,
            github_url=None,
            github_token=None,
            github_username=None,
            github_password=None):

        self.base_path = base_path
        self.repo_owner = repo_owner
        self.repo_name = repo_name
        self.repo_ref = repo_ref or 'master'
        self.github_url = github_url
        self.github_token = github_token
        self.github_username = github_username
        self.github_password = github_password

        self.github = self._get_github()
        self.repository = self._get_repository()

    def _get_github(self):
        try:
            import github3
        except ImportError:
            raise Exception("""
            ERROR: github3.py not installed!  Please install via
              pip install boundary-layer[github]
            and try again.""")

        if self.github_url:
            return github3.GitHubEnterprise(
                url=self.github_url,
                username=self.github_username,
                password=self.github_password,
                token=self.github_token)

        return github3.GitHub(
            username=self.github_username,
            password=self.github_password,
            token=self.github_token)

    def _get_repository(self):
        return self.github.repository(
            self.repo_owner,
            self.repo_name)

    def fetch_file_content(self, rel_path):
        full_path = os.path.join(self.base_path, rel_path.lstrip('/'))
        content_b64 = self.repository.file_contents(
            path=full_path,
            ref=self.repo_ref).content

        return base64.b64decode(content_b64)



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import os
import xmltodict
import networkx as nx
import six

from boundary_layer.graph import _GraphUtil
from boundary_layer.logger import logger
from boundary_layer import exceptions, VERSION_STRING, plugins
from boundary_layer.oozier.jsp_macros import JspMacroTranslator
from boundary_layer.oozier.schema import OozieWorkflowSchema
from boundary_layer.oozier.actions import OozieSubWorkflowBuilderBase
from boundary_layer.workflow import Workflow


class OozieWorkflowParser(object):
    def __init__(
            self,
            file_fetcher,
            prune_forks=False,
            prune_joins=False,
            production=False,
            debug=False,
    ):

        self.file_fetcher = file_fetcher
        self.prune_forks = prune_forks
        self.prune_joins = prune_joins
        self.production = production
        self.debug = debug

    def load_workflow(self, primary_workflow_name, args):
        oozie_config = plugins.manager.get_oozie_config(args)

        cluster_config = oozie_config.cluster_config()

        (primary_dag, sub_dags) = self._parse_and_build_dags(
            primary_workflow_name,
            cluster_config,
            oozie_config)

        if self._requires_cluster_resource(primary_dag, sub_dags, cluster_config):
            primary_dag = self._insert_cluster_resource(
                primary_dag, cluster_config)

        primary_dag['compatibility_version'] = VERSION_STRING

        if oozie_config.plugin_config():
            primary_dag['plugin_config'] = oozie_config.plugin_config()

        if oozie_config.default_task_args():
            primary_dag['default_task_args'] = oozie_config.default_task_args()

        if oozie_config.dag_args():
            primary_dag['dag_args'] = oozie_config.dag_args()

        if oozie_config.dag_imports():
            primary_dag['imports'] = oozie_config.dag_imports()

        primary_dag['name'] = args.dag_name or primary_dag['name']

        # Build the workflow
        wf = Workflow(
            primary=primary_dag,
            secondary=list(sub_dags.values()),
            filename='oozie:' + primary_workflow_name,
        )

        # Prune away any unwanted nodes
        if args.prune_nodes or args.only_nodes:
            wf = wf.prune(
                prune_nodes=args.prune_nodes,
                only_nodes=args.only_nodes)

        # Now add in any additional operators that we require.  We have to do
        # this down here in order to avoid having these nodes pruned out above.
        if oozie_config.upstream_operators():
            primary_copy = wf.primary.copy()
            primary_copy['before'] = oozie_config.upstream_operators()
            wf = Workflow(
                primary=primary_copy,
                secondary=wf.secondary,
                filename='oozie:' + primary_workflow_name)

        return wf

    @staticmethod
    def partition_actions(data):
        operators = [ac for ac in data['action'] if not isinstance(ac, OozieSubWorkflowBuilderBase)]
        sub_workflows = [ac for ac in data['action'] if isinstance(ac, OozieSubWorkflowBuilderBase)]

        logger.debug('Partitioned actions for workflow %s.  Operators are %s, sub_workflows are %s',
                     data['name'],
                     [operator.name for operator in operators],
                     [sub_workflow.name for sub_workflow in sub_workflows])

        assert len(operators + sub_workflows) == len(data['action']), \
            'Some actions were not partitioned!'

        return {
            'operators': operators,
            'sub_workflows': sub_workflows,
        }

    def _parse_workflow(self, filename, cluster_config, oozie_config):
        parsed = xmltodict.parse(
            self.file_fetcher.fetch_file_content(filename))

        loaded = OozieWorkflowSchema(context={
            'cluster_config': cluster_config,
            'oozie_plugin': oozie_config,
            'macro_translator': JspMacroTranslator(oozie_config.jsp_macros()),
            'production': self.production,
        }).load(parsed)

        if loaded.errors:
            raise Exception('Errors parsing file {}: {}'.format(
                filename,
                loaded.errors))

        data_copy = loaded.data.copy()
        data_copy.update(self.partition_actions(loaded.data))

        return data_copy

    @staticmethod
    def _sub_workflow_target_name(sub_workflow_action):
        return '/'.join(sub_workflow_action['app_path'].split('/')[3:])

    def _parse_all(
            self,
            primary_workflow_path_name,
            cluster_config,
            oozie_config):

        def _build_workflow_path(name):
            return os.path.join(name, 'workflow.xml')

        path = _build_workflow_path(primary_workflow_path_name)
        logger.debug('parsing primary workflow from path %s', path)
        primary = self._parse_workflow(path, cluster_config, oozie_config)

        sub_workflows = {}
        parsed_sub_workflow_targets = set()
        q = [wf for wf in primary['sub_workflows']]

        while q:
            item = q.pop()
            swf_path_name = self._sub_workflow_target_name(item.get_action())
            if swf_path_name in parsed_sub_workflow_targets:
                continue

            parsed_sub_workflow_targets.add(swf_path_name)

            swf_path = _build_workflow_path(swf_path_name)

            logger.debug('parsing sub workflow from path %s', swf_path)
            swf = self._parse_workflow(
                swf_path, cluster_config, oozie_config)

            sub_workflows[swf['name']] = swf
            item.set_target_name(swf['name'])

            q += [wf for wf in swf['sub_workflows']]

        return (primary, sub_workflows)

    def _flow_control_nodes(self, wf):
        result = {
            'start': ([wf['start']['to']], wf['start']),
            wf['end'].get('name', 'end'): ([], wf['end'])
        }

        if wf.get('kill'):
            result[wf['kill'].get('name', 'kill')] = ([], wf['kill'])

        for decision in wf['decision']:
            if not self.debug:
                raise Exception('decision node found: `{}` not supported right '
                                'now!  Set --debug to build without raising this error, but '
                                'only at your own risk!'.format(decision['name']))

            logger.warning(
                'decision node found: `%s`.  Including all downstream '
                'branches because --debug was specified', decision['name'])

            cases = [case['to'] for case in decision['switch']['case']]
            default = decision['switch']['default']['to']
            result[decision['name']] = (cases + [default], decision)

        return result

    @staticmethod
    def _fork_nodes(wf):
        return {
            fork['name']: ([path['start'] for path in fork['path']], fork)
            for fork in wf['fork']
        }

    @staticmethod
    def _join_nodes(wf):
        return {join['name']: ([join['to']], join) for join in wf['join']}

    @staticmethod
    def _operator_nodes(wf):
        return {op.name: ([op.ok_to], op) for op in wf['operators']}

    @staticmethod
    def _sub_workflow_nodes(wf):
        return {swf.name: ([swf.ok_to], swf) for swf in wf['sub_workflows']}

    def _all_nodes(self, wf):
        result = {}
        result.update(self._flow_control_nodes(wf))
        result.update(self._fork_nodes(wf))
        result.update(self._join_nodes(wf))
        result.update(self._operator_nodes(wf))
        result.update(self._sub_workflow_nodes(wf))

        return result

    @staticmethod
    def _build_graph(graph_name, nodes):
        graph = nx.DiGraph()

        logger.debug('Creating graph %s from nodes %s',
                     graph_name, nodes.keys())

        graph.add_nodes_from(nodes.keys())

        graph.add_edges_from(
            (upstream_name, downstream_name)
            for (upstream_name, (downstream_node_names, node)) in six.iteritems(nodes)
            for downstream_name in downstream_node_names)

        if not nx.algorithms.components.is_weakly_connected(graph):
            components = list(
                nx.algorithms.components.weakly_connected_components(graph))
            logger.warning(
                'Multiple connected components found for graph `%s`: %s',
                graph_name,
                components)

        if not nx.algorithms.dag.is_directed_acyclic_graph(graph):
            raise exceptions.CyclicWorkflowException(
                'Invalid graph `{}`: not a DAG!'.format(graph_name))

        logger.debug('Successfully created graph %s with nodes %s',
                     graph_name, graph.nodes)

        return graph

    def _prune_flow_control_nodes(self, wf, graph):
        prune_nodes = list(self._flow_control_nodes(wf).keys())

        if self.prune_forks:
            prune_nodes += list(self._fork_nodes(wf).keys())

        if self.prune_joins:
            prune_nodes += list(self._join_nodes(wf).keys())

        return _GraphUtil.prune_nodes(graph, nodes=prune_nodes)

    def _build_dag(self, wf):
        graph = self._build_graph(wf['name'], self._all_nodes(wf))

        self._prune_flow_control_nodes(wf, graph)

        operator_nodes = self._operator_nodes(wf)

        fork_join_nodes = self._fork_nodes(wf)
        fork_join_nodes.update(self._join_nodes(wf))

        sub_workflow_nodes = self._sub_workflow_nodes(wf)

        operators = []
        sub_workflows = []

        # Build in dependency-resolved order, for easier viewing by humans
        for node_name in nx.algorithms.dag.topological_sort(graph):
            upstream = list(graph.predecessors(node_name))
            base_operator = {
                'upstream_dependencies': upstream} if upstream else {}

            node = fork_join_nodes.get(node_name)
            if node:
                base_operator.update(node[1]['operator'])
                base_operator['name'] = node[1]['name']  # must take the oozie action name
                operators.append(base_operator)
                continue

            node = operator_nodes.get(node_name)
            if node:
                base_operator.update(node[1].get_operator())
                operators.append(base_operator)
                continue

            swf = sub_workflow_nodes.get(node_name)
            if not swf:
                raise Exception('Unrecognized node name: `{}` (swf nodes are {})'.format(
                    node_name, list(sub_workflow_nodes.keys())))

            base_operator.update(swf[1].get_operator())

            sub_workflows.append(base_operator)

        result = {
            'name': wf['name'],
        }

        if operators:
            result['operators'] = operators

        if sub_workflows:
            result['sub_dags'] = sub_workflows

        return result

    def _parse_and_build_dags(
            self,
            primary_workflow_name,
            cluster_config,
            oozie_plugin):

        (primary, sub) = self._parse_all(
            primary_workflow_name,
            cluster_config,
            oozie_plugin)

        primary_dag = self._build_dag(primary)

        sub_dags = {}
        for (name, swf) in six.iteritems(sub):
            sub_dags[name] = self._build_dag(swf)

        return self._fixup_dags(primary_dag, sub_dags, cluster_config)

    def _fixup_dags(self, primary_dag, sub_dags, cluster_config):
        fixedup_primary_dag = self._fixup_subdags(
            primary_dag, sub_dags, cluster_config)

        # re-key the sub_dags by their actual name, rather than by their
        # path name
        fixedup_sub_dags = {
            dag['name']: dag for dag in
            [self._fixup_subdags(sub_dag, sub_dags, cluster_config)
             for sub_dag in sub_dags.values()]
        }

        return (fixedup_primary_dag, fixedup_sub_dags)

    def _fixup_subdags(self, dag, all_sub_dags, cluster_config):
        """
            Once we have loaded the primary dag and all of the sub-dags in
            isolation, we have to fill-in some values that can only be
            determined collectively.  For now, this includes only:
            1. We must make sure that any resources required by a subdag are
               passed to it from the parent dag (e.g. a dataproc cluster that
               is created by the primary but used to run jobs in the subdag)
        """
        subdags = dag.get('sub_dags')
        if not subdags:
            return dag

        copy = dag.copy()
        copy['sub_dags'] = [sd.copy() for sd in subdags]
        for sd in copy['sub_dags']:
            target_subdag = all_sub_dags.get(sd['target'])
            if self._requires_cluster_resource(target_subdag, all_sub_dags, cluster_config):
                sd['requires_resources'] = [cluster_config.managed_resource['name']]

        return copy

    def _requires_cluster_resource(self, dag, sub_dags, cluster_config):
        if not cluster_config.managed_resource:
            return False

        resource_name = cluster_config.managed_resource['name']
        return any(resource_name in op.get('requires_resources', [])
                   for op in dag.get('operators', [])) or \
            any(self._requires_cluster_resource(sub_dags[sd['target']], sub_dags, cluster_config)
                for sd in dag.get('sub_dags', []))

    @staticmethod
    def _insert_cluster_resource(wf, cluster_config):
        copy = wf.copy()
        copy['resources'] = [cluster_config.managed_resource]

        return copy



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import operator
import pkg_resources
import marshmallow as ma
from boundary_layer.plugins.oozie_plugin import OozieMetaPlugin
from boundary_layer.logger import logger


class PluginManager(object):
    ENTRY_POINT_GROUP = 'boundary_layer_plugins'

    def __init__(
            self,
            load_package_plugins=True,
            plugins=None):

        self._plugins = [p for p in (plugins or [])]

        if load_package_plugins:
            self._plugins += self._load_package_plugins()

        logger.info('Loaded plugins %s', ', '.join(
            plugin.name for plugin in self._plugins))

    def validate_config(self, plugin_config):
        plugin_names = frozenset(plugin.name for plugin in self._plugins)

        bad_plugin_names = [
            name for name in plugin_config if name not in plugin_names]
        if bad_plugin_names:
            raise Exception(
                'Unrecognized plugin name(s) `{}` in plugin config; perhaps '
                'these plugins are not installed?'.format(
                    '`, `'.join(bad_plugin_names)))

        for plugin in self._plugins:
            self.parse_plugin_config(plugin, plugin_config.get(plugin.name, {}))

    def parse_plugin_config(self, plugin, config):
        if not plugin.config_schema_cls and config:
            raise Exception(
                'Config provided for plugin {}, but this plugin does not have '
                'a config schema!'.format(plugin.name))

        if not plugin.config_schema_cls:
            return None

        if not issubclass(plugin.config_schema_cls, ma.Schema):
            raise Exception(
                'Config schema for plugin {} is not a marshmallow Schema. '
                'Found: {}'.format(plugin.name, plugin.config_schema_cls))

        parsed_config = plugin.config_schema_cls().load(config or {})
        if parsed_config.errors:
            raise Exception(
                'Errors parsing configuration for plugin {}: {}'.format(
                    plugin.name,
                    parsed_config.errors))

        return parsed_config.data

    def insert_imports(self, plugin_config):
        objects = []
        modules = []

        plugin_config = plugin_config or {}
        for plugin in self._plugins:
            conf = self.parse_plugin_config(plugin, plugin_config.get(plugin.name))
            plugin_imports = plugin.insert_imports(conf) or {}

            objects += plugin_imports.get('objects', [])
            modules += plugin_imports.get('modules', [])

        result = {}
        if objects:
            result['objects'] = objects

        if modules:
            result['modules'] = modules

        return result

    def _build_insert_dict(self, plugin_config, method):
        result = {}
        plugin_config = plugin_config or {}
        for plugin in self._plugins:
            conf = self.parse_plugin_config(plugin, plugin_config.get(plugin.name))
            plugin_items = method(plugin)(conf)

            result.update(plugin_items or {})

        return result

    def insert_default_task_args(self, plugin_config):
        return self._build_insert_dict(
            plugin_config,
            operator.attrgetter('insert_default_task_args'))

    def insert_dag_args(self, plugin_config):
        return self._build_insert_dict(
            plugin_config,
            operator.attrgetter('insert_dag_args'))

    def _build_insert_list(self, plugin_config, method):
        plugin_config = plugin_config or {}
        items = []
        for plugin in self._plugins:
            conf = self.parse_plugin_config(plugin, plugin_config.get(plugin.name))
            plugin_items = method(plugin)(conf)

            items += (plugin_items or [])

        return items

    def insert_before(self, plugin_config):
        return self._build_insert_list(plugin_config, operator.attrgetter('insert_before'))

    def insert_operators(self, plugin_config):
        return self._build_insert_list(plugin_config, operator.attrgetter('insert_operators'))

    def insert_after(self, plugin_config):
        return self._build_insert_list(plugin_config, operator.attrgetter('insert_after'))

    def operators(self, item):
        return self._registry_lookup(item, 'operator')

    def generators(self, item):
        return self._registry_lookup(item, 'generator')

    def subdags(self, item):
        return self._registry_lookup(item, 'subdag')

    def resources(self, item):
        return self._registry_lookup(item, 'resource')

    def property_preprocessors(self, item):
        return self._registry_lookup(item, 'property_preprocessor')

    def get_oozie_config(self, args):
        return OozieMetaPlugin(self._plugins, args)

    def _get_oozie_plugin_classes(self):
        return {
            plugin.name: plugin.oozie_plugin_cls for plugin in self._plugins
            if plugin.oozie_plugin_cls
        }

    def register_oozie_parser_arguments(self, parser):
        for cls in self._get_oozie_plugin_classes().values():
            cls.register_arguments(parser)

    def _load_package_plugins(self):
        return [entry_point.load()() for entry_point in
                pkg_resources.iter_entry_points(self.ENTRY_POINT_GROUP)]

    def _registry_lookup(self, item, registry_name):
        registry_getter = operator.attrgetter(registry_name + '_registry')

        prioritized_hits = {}
        for plugin in self._plugins:
            registry = registry_getter(plugin)
            if not registry:
                continue

            hit = registry.get(item)
            if not hit:
                continue

            prioritized_hits.setdefault(plugin.priority, [])
            prioritized_hits[plugin.priority].append(hit)

        if not prioritized_hits:
            available_types = frozenset(
                node_type for plugin in self._plugins
                if registry_getter(plugin)
                for node_type in registry_getter(plugin).node_configs)
            raise Exception(
                'Lookup failed for registry `{}`: `{}` not found for item `{}`. '
                'Available item types are: [`{}`]'.format(
                    registry_name,
                    item['type'],
                    item,
                    '`, `'.join(available_types)))

        highest_hit_priority = max(
            prioritized_hits.keys(),
            key=operator.attrgetter('value'))

        if not highest_hit_priority:
            raise Exception('Unknown {}: {}'.format(registry_name, item))

        hits = prioritized_hits[highest_hit_priority]

        if len(hits) > 1:
            raise Exception(
                'Ambiguous {registry_name} specification: {item} matches on '
                'multiple known {registry_name}s for plugins with priority '
                '{priority}: {hits}'.format(
                    registry_name=registry_name,
                    item=item,
                    priority=highest_hit_priority,
                    hits=hits))

        return hits[0]



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import abc
import os
from enum import Enum
from boundary_layer.logger import logger
from boundary_layer.plugins.oozie_plugin import BaseOozieParserPlugin

from boundary_layer.registry.types import \
    OperatorRegistry, \
    SubdagRegistry, \
    GeneratorRegistry, \
    ResourceRegistry, \
    PropertyPreprocessorRegistry


class PluginPriority(Enum):
    FALLBACK = -1  # If all else fails...
    DEFAULT = 0  # The default plugin only, most likely
    SHARED = 1  # Can be overridden, but can override default
    FINAL = 2  # Cannot be overridden


class BasePlugin(object):
    __metaclass__ = abc.ABCMeta

    priority = PluginPriority.SHARED

    property_preprocessors = []

    def __init__(self):
        logger.debug('Loading plugin %s', self.name)

        if not isinstance(self.oozie_plugin_cls, type(None)) and not \
                issubclass(self.oozie_plugin_cls, BaseOozieParserPlugin):
            raise Exception('Invalid oozie plugin: {}'.format(
                self.oozie_plugin_cls))

        self.operator_registry = self.load_config_registry(
            'operators', OperatorRegistry)
        self.subdag_registry = self.load_config_registry('subdags', SubdagRegistry)
        self.generator_registry = self.load_config_registry(
            'generators', GeneratorRegistry)
        self.resource_registry = self.load_config_registry(
            'resources', ResourceRegistry)
        self.property_preprocessor_registry = \
            PropertyPreprocessorRegistry(self.property_preprocessors)

    @abc.abstractproperty
    def name(self):
        pass

    configs_path = None

    # Overrideable attribute for specifying an oozie parser plugin class
    # from the main plugin instance
    oozie_plugin_cls = None

    # Overrideable attribute for specifying a schema to use to parse this
    # plugin's entry in the plugin_config field of a DAG.  If set to None,
    # then an error will be raised if there are any entries in plugin_config
    # for this plugin.
    config_schema_cls = None

    def insert_imports(self, config):
        pass

    def insert_default_task_args(self, config):
        pass

    def insert_dag_args(self, config):
        pass

    def insert_before(self, config):
        pass

    def insert_operators(self, config):
        pass

    def insert_after(self, config):
        pass

    @classmethod
    def load_config_registry(cls, subpath, factory):
        logger.debug('Configs path: %s', cls.configs_path)
        if not cls.configs_path:
            return None

        registry_config_path = os.path.join(cls.configs_path, subpath)
        if not os.path.isdir(registry_config_path):
            return None

        return factory([registry_config_path])



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re
import operator
import hashlib
from six.moves import map
import six
from boundary_layer.containers import ExecutionContext


class GenericNamedParameterPasser(object):

    def __init__(self, typ, args):
        self.typ = typ
        self.args = args

    def __repr__(self):
        return '{}({})'.format(self.typ,
                               ",".join([k + "=" + repr(v) for k, v in self.args.items()]))


def freeze(item):
    if isinstance(item, dict):
        return frozenset((key, freeze(value)) for (key, value) in six.iteritems(item))

    if isinstance(item, (list, tuple)):
        return tuple(map(freeze, item))

    if isinstance(item, set):
        return frozenset(map(freeze, item))

    return item


def md5(item):
    digest = hashlib.md5()

    digest.update(item.encode('utf-8'))

    return digest.hexdigest()


def make_flow_control_node(deps, name=None, default_task_args=None):
    """ Make a flow control node.
    """
    from boundary_layer import plugins

    sorted_deps = sorted(deps, key=operator.attrgetter('name'))

    def build_hash():
        return md5('-'.join(dep.name for dep in sorted_deps))

    def name_from_deps():
        max_name_len = 60
        hash_len = 6
        trailing_len = len('-fc')
        chars_per_dep = -1 + (max_name_len - hash_len -
                              trailing_len) // len(deps)

        return '-'.join([dep.name[:chars_per_dep] for dep in sorted_deps] +
                        [build_hash()[:hash_len], 'fc'])

    fc_node = plugins.manager.operators({
        'name': name or name_from_deps(),
        'type': 'flow_control',
        'upstream_dependencies': list(deps),
    })
    fc_node.resolve_properties(
        default_task_args=default_task_args,
        execution_context=ExecutionContext(referrer=None, resources={}),
        base_operator_loader=plugins.manager.operators,
        preprocessor_loader=plugins.manager.property_preprocessors)

    return fc_node


def merge_schemas(parent, child):
    """ Merge two json schemas.  Note that this method is asymmetric;
        if the child redefines a field present in the parent, the child's
        definition takes precedence.  But if either of them provide
        fields in the `requires` directive, those will be merged.

        :param parent: the parent operator's json schema
        :type parent: dict
        :param child: the child operator's json schema
        :type child: dict
        :returns: a new jsonschema reflecting the merger
        :rtype: dict
    """
    result = {'properties': {}}
    result['properties'].update(parent.get('properties', {}))
    result['properties'].update(child.get('properties', {}))

    required = parent.get('required', []) + child.get('required', [])
    if required:
        result['required'] = required

    for schema in [parent, child]:
        if 'additionalProperties' in schema:
            result['additionalProperties'] = schema['additionalProperties']

    extra_keys = (frozenset(parent) | frozenset(child)) - \
        frozenset(['properties', 'required', 'additionalProperties'])

    if extra_keys:
        raise Exception(
            'Cannot merge schemas due to unrecognized fields: {} '
            '(parent schema: {}; child schema: {})'.format(
                extra_keys,
                parent,
                child))

    return result


def sanitize_operator_name(name):
    # make operator names snake_case
    if isinstance(name, list):
        return list(map(sanitize_operator_name, name))
    elif isinstance(name, six.string_types):
        assert name[0].isalpha() or name[0] == '_', \
            'Invalid name `{}`: must start with a character!'.format(name)

        return re.sub('[^a-zA-Z0-9_]', '_', name)

    raise Exception('Cannot sanitize name `{}`'.format(name))



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from boundary_layer.registry import ConfigFileRegistry, RegistryNode, NodeTypes
from boundary_layer.schemas.internal.resources import ResourceSpecSchema


class ResourceNode(RegistryNode):
    type = NodeTypes.RESOURCE

    def imports(self):
        create = self.create_operator.imports()
        destroy = self.destroy_operator.imports()

        return {
            'modules': create['modules'] + destroy['modules'],
            'objects': create['objects'] + destroy['objects'],
        }

    def __init__(self, config, item):
        super(ResourceNode, self).__init__(config, item)
        self._create_operator_node = None
        self._destroy_operator_node = None

    @property
    def create_operator_type(self):
        return self.config['create_operator_type']

    @property
    def create_operator(self):
        if not self._create_operator_node:
            raise Exception(
                'resolve_properties() has not yet been called on {}'.format(
                    self))
        return self._create_operator_node

    @property
    def destroy_operator_type(self):
        return self.config.get('destroy_operator_type')

    @property
    def disable_sentinel_node(self):
        """ Determine whether the DAG or the resource config specifies that
            we should disable the sentinel node (or not).  Note that we allow
            the DAG to override any configuration that may be present in the
            resource config, rather than vice-versa.
        """
        return self.item.get(
            'disable_sentinel_node',
            self.config.get('disable_sentinel_node', False))

    @property
    def destroy_operator(self):
        if not self.destroy_operator_type:
            return None

        if not self._destroy_operator_node:
            raise Exception(
                'resolve_properties() has not yet been called on {}'.format(
                    self))
        return self._destroy_operator_node

    @property
    def provides_args(self):
        return self.config['provides_args']

    def resolve_properties(
            self,
            execution_context,
            default_task_args,
            base_operator_loader,
            preprocessor_loader):

        self._create_operator_node = base_operator_loader({
            'name': self.name + '_create',
            'type': self.create_operator_type,
            'properties': self.item.get('properties', {}),
            'upstream_dependencies': self.upstream_dependencies,
            'downstream_dependencies': self.downstream_dependencies,
            'requires_resources': self.requires_resources,
        })

        create_properties = self._create_operator_node.resolve_properties(
            execution_context=execution_context,
            default_task_args=default_task_args,
            base_operator_loader=base_operator_loader,
            preprocessor_loader=preprocessor_loader)

        if not self.destroy_operator_type:
            return (create_properties, None)

        self._destroy_operator_node = base_operator_loader({
            'name': self.name + '_destroy',
            'type': self.destroy_operator_type,
            'properties': {},
            'requires_resources': [self.name]
        })

        resource_context_with_self = execution_context.resources.copy()
        resource_context_with_self[self.name] = self
        destroy_properties = self._destroy_operator_node.resolve_properties(
            execution_context=execution_context._replace(resources=resource_context_with_self),
            default_task_args=default_task_args,
            base_operator_loader=base_operator_loader,
            preprocessor_loader=preprocessor_loader)

        return (create_properties, destroy_properties)

    def get_provided_args(self):
        create_operator_properties = self.create_operator.resolved_properties
        return {
            name: create_operator_properties.values.get(name)
            for name in self.provides_args
        }


class ResourceRegistry(ConfigFileRegistry):
    spec_schema_cls = ResourceSpecSchema

    node_cls = ResourceNode



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import marshmallow as ma

# Define a base schema type that rejects any inputs with unrecognized arguments,
# to prevent silent mis-configurations.  Based on the docs here:
# http://marshmallow.readthedocs.io/en/latest/extending.html


class StrictSchema(ma.Schema):
    @ma.validates_schema(pass_original=True)
    def check_no_unknowns(self, _, original_data):
        def check_single_datum(datum):
            unknown = set(datum) - set(self.fields)
            if unknown:
                raise ma.ValidationError('Unknown fields present', unknown)

        if self.many:
            for item in original_data:
                check_single_datum(item)
        else:
            check_single_datum(original_data)



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import functools
from six.moves import filter
import networkx as nx
from boundary_layer.logger import logger
from boundary_layer.exceptions import CyclicWorkflowException, InvalidFlowControlNode
from boundary_layer import util


class _GraphUtil(object):
    @staticmethod
    def get_upstream_surface(graph):
        return [node for node in graph.nodes()
                if not list(graph.predecessors(node))]

    @staticmethod
    def get_downstream_surface(graph):
        return [node for node in graph.nodes()
                if not list(graph.successors(node))]

    @staticmethod
    def upstream_dependency_set(node, graph):
        return frozenset(graph.predecessors(node))

    @staticmethod
    def downstream_dependency_set(node, graph):
        return frozenset(graph.successors(node))

    @staticmethod
    def leaf_nodes(graph):
        return [node for node in graph.nodes()
                if not list(graph.successors(node))]

    @staticmethod
    def get_distinct_upstream_dependency_sets(nodes, graph):
        dependency_sets = [
            _GraphUtil.upstream_dependency_set(node, graph)
            for node in nodes]

        return frozenset(filter(None, dependency_sets))

    @staticmethod
    def build_subgraph(nodes):
        graph = nx.DiGraph()

        node_lookup = {node.name: node for node in nodes}

        # First insert every node into the graph
        graph.add_nodes_from(nodes)

        # Now insert every explicit edge into the graph, making sure that
        # any referenced node is actually present in the subgraph
        for node in nodes:
            logger.debug('Processing node %s: upstream %s downstream %s',
                         node, node.upstream_dependencies, node.downstream_dependencies)
            for upstream in node.upstream_dependencies:
                upstream_node = node_lookup.get(upstream)
                if not upstream_node:
                    raise Exception('Invalid reference {} <- {}: perhaps it '
                                    'crosses sub-graph boundaries?'.format(
                                        upstream,
                                        node.name))

                graph.add_edge(upstream_node, node)

            for downstream in node.downstream_dependencies:
                downstream_node = node_lookup.get(downstream)

                if not downstream_node:
                    raise Exception('Invalid reference {} -> {}: perhaps it '
                                    'crosses sub-graph boundaries?'.format(
                                        node.name,
                                        downstream))

                graph.add_edge(node, downstream_node)

        if not nx.algorithms.dag.is_directed_acyclic_graph(graph):
            raise CyclicWorkflowException('Invalid sub-graph: not a DAG!')

        return graph

    @staticmethod
    def ordered(graph):
        return list(nx.algorithms.dag.topological_sort(graph))

    @staticmethod
    def attach_flow_control_between(
            upstream,
            downstream,
            graph,
            fc_node=None,
            fc_node_builder=None):
        """ For each downstream node, break that node's connections with all of
            the upstream nodes provided, and insert a flow-control node between
            them.  If the fc_node argument is provided, then use the provided node
            for flow-control; otherwise, insert a new flow-control node.
        """
        if not upstream or not downstream:
            raise Exception('upstream and downstream node lists cannot be '
                            'empty (given upstream `{}`, downstream `{}`)'.format(
                                upstream, downstream))

        if not fc_node and not fc_node_builder:
            raise Exception(
                'Internal error: neither fc_node nor fc_node_builder argument '
                'was provided to the method attach_flow_control_between() '
                'but exactly one of these is required')

        if fc_node and fc_node_builder:
            raise Exception(
                'Internal error: both fc_node and fc_node_builder arguments '
                'were provided to the method attach_flow_control_between() '
                'but exactly one of these is required')

        if not fc_node:
            fc_node = fc_node_builder(deps=downstream)
            logger.debug('Inserting new flow control node %s between '
                         '{ %s } and { %s }', fc_node, upstream, downstream)

            graph.add_node(fc_node)
        else:
            logger.debug('Attaching existing node %s between '
                         '{ %s } and { %s }', fc_node, upstream, downstream)

        for node in downstream:
            if not upstream.issubset(_GraphUtil.upstream_dependency_set(node, graph)):
                raise InvalidFlowControlNode(
                    'Internal error: Invalid flow-control insertion. '
                    'Upstream nodes {} are not all upstream dependencies of '
                    'node {}'.format(upstream, node))

        # Break the edges between the upstream and downstream nodes
        break_edges = [(up, down) for up in upstream for down in downstream]
        logger.debug('removing edges: %s', break_edges)
        graph.remove_edges_from(break_edges)

        # Now insert the flow control node between the upstream and downstream
        # dependencies
        add_edges = [(up, fc_node) for up in upstream] + \
                    [(fc_node, down) for down in downstream]

        logger.debug('adding edges: %s', add_edges)
        graph.add_edges_from(add_edges)

    @staticmethod
    def upstream_resource_boundary(resource_name, graph):
        return [node for node in graph.nodes()
                if resource_name in node.requires_resources and
                all(resource_name not in ancestor.requires_resources
                    for ancestor in nx.algorithms.dag.ancestors(graph, node))]

    @staticmethod
    def downstream_resource_boundary(resource_name, graph):
        """ Determine the downstream boundary nodes for the provided resource.
            This is the set of nodes N where, for any n in N, n requires the
            resource, but there do not exist any m, downstream of n, such that
            m also requires the resouce

            :param resource_name: name of the resource
            :type resource_name: string
            :param graph: the graph
            :type graph: nx.DiGraph
            :return The downstream boundary for the provided resource
            :rtype list<boundary_layer.registry.RegistryNode>
        """
        return [node for node in graph.nodes()
                if resource_name in node.requires_resources and
                all(resource_name not in descendant.requires_resources
                    for descendant in nx.algorithms.dag.descendants(graph, node))]

    @staticmethod
    def requires_destroy_resource_sentinel(downstream_resource_boundary, graph):
        """ Determine whether the destroy-resource operator created downstream
            of the specified resource will mask any upstream task failures, and
            therefore whether we should insert a sentinel node in parallel with
            the destroy-resource node.  This is true if any of the downstream
            resource boundary nodes are leaf nodes of the graph, ie nodes that
            have no downstream nodes to which to propagate error statuses.

            :param downstream_resource_boundary: downstream resource boundary
            :type downstream_resource_boundary: list<boundary_layer.registry.RegistryNode>
            :param graph: the graph
            :type graph: nx.DiGraph
        """
        return bool(
            frozenset(_GraphUtil.leaf_nodes(graph)) &
            frozenset(downstream_resource_boundary))

    @staticmethod
    def insert_bundler_nodes(nodes, graph, fc_node_builder, default_bundler_node=None):
        """ Utility for inserting "bundler" nodes that help us express
            complex dependency settings that are not otherwise supported by
            Airflow.  In particular: suppose we have a node, N, that we would
            like to trigger whenever either (A and B) are complete or
            (C and D) are complete.  Airflow does not have any mechanism for
            expressing this directly.  Instead, we insert a bundler node X0
            that triggers when (A and B) are complete, and a bundler node X1
            that triggers when (C and D) are complete, and then we set Z to
            trigger when either X0 or X1 is complete, using the ONE_SUCCESS
            trigger rule.

            The specific use case for this logic in boundary-layer is for
            creating shared resources when any resource-dependent operator is
            ready, based on its upstream dependencies being satisfied, even if
            the upstream dependencies of other resource-dependent operators
            may not yet be satisfied.

            :param bundles: list of sets of nodes
            :param graph: graph
            :param default_bundler_node: preexisting node to use as the bundler
                if there is ony one bundle in the input.  This parameter is
                ignored if there are more than one bundle; in that case, or if
                this parameter is not provided at all, new flow control nodes
                will be inserted to act as bundler nodes.
        """

        bundles = _GraphUtil.get_distinct_upstream_dependency_sets(
            nodes, graph)

        bundler_nodes = _GraphUtil._create_bundler_nodes(
            bundles,
            fc_node_builder,
            default_bundler_node,
            )

        to_insert = [node for node in bundler_nodes.values() if node !=
                     default_bundler_node]
        logger.debug('Inserting bundler nodes: %s', to_insert)
        graph.add_nodes_from(to_insert)

        for node in nodes:
            dep_set = _GraphUtil.upstream_dependency_set(node, graph)

            if not dep_set:
                # Some nodes may not have had any upstream dependencies, and
                # no bundler nodes would have been added for these
                continue

            bundler_node = bundler_nodes[dep_set]

            _GraphUtil.attach_flow_control_between(
                upstream=dep_set,
                downstream=frozenset([node]),
                graph=graph,
                fc_node=bundler_node)

        return bundler_nodes

    @staticmethod
    def _create_bundler_nodes(bundles, fc_node_builder, default_bundler_node=None):
        if len(bundles) == 1 and default_bundler_node:
            return {
                list(bundles)[0]: default_bundler_node,
            }

        # If there is more than one way in which the boundary nodes
        # depend on upstream nodes, then we create flow-control
        # nodes that allow us to trigger the resource creation
        # whenever the first boundary node would have otherwise
        # been available (even if others still have unmet dependencies).
        # Note that we may opt not to do this, and instead trigger only
        # after all of the boundary nodes would have been ready, but that
        # gets pushed to a higher level configuration.
        return {
            deps: fc_node_builder(deps=deps)
            for deps in bundles
        }

    @staticmethod
    def attach_create_resource(resource, graph, fc_node_builder):
        create_resource_node = resource.create_operator
        graph.add_node(create_resource_node)

        upstream_boundary = \
            _GraphUtil.upstream_resource_boundary(resource.name, graph)

        logger.debug('upstream boundary nodes for resource `%s`: %s',
                     resource.name, upstream_boundary)

        bundler_nodes = _GraphUtil.insert_bundler_nodes(
            upstream_boundary,
            graph,
            fc_node_builder,
            create_resource_node)

        for bundler_node in bundler_nodes.values():
            if bundler_node == create_resource_node:
                continue

            graph.add_edge(bundler_node, create_resource_node)

        for node in upstream_boundary:
            dep_set = _GraphUtil.upstream_dependency_set(node, graph)
            assert len(dep_set) <= 1, \
                'Error: node {} has dependency set {}, but its ' \
                'dependencies should have been bundled!'.format(
                    node,
                    dep_set)

            bundler_node = list(dep_set)[0] if dep_set else None

            # If the create_resource_node is functioning as the bundler
            # node, then all necessary dependencies will have been set up
            # already.
            if bundler_node == create_resource_node:
                continue

            # Otherwise, we have to insert dependencies between the
            # create_resource_node and the downstream node.
            if not bundler_node:
                logger.debug(
                    'Adding edge between create_resource_node and dependency-free node %s',
                    node)
            else:
                logger.debug(
                    'Adding edge between create_resource_node and node %s with bundler node %s',
                    node,
                    bundler_node)

            graph.add_edge(create_resource_node, node)

    @staticmethod
    def attach_destroy_resource(resource, graph, fc_node_builder):
        """ Method to attach the destroy-resource node to the graph,
            which is simpler than the attachment of the create-resource node
            because it can be attached directly to the downstream boundary nodes
            without breaking any existing dependencies.
        """
        destroy_resource = resource.destroy_operator
        if not destroy_resource:
            return

        downstream_boundary = _GraphUtil.downstream_resource_boundary(
            resource.name,
            graph)

        logger.debug('downstream boundary nodes for resource `%s`: %s',
                     resource.name, downstream_boundary)

        nodes_to_add = [destroy_resource]

        if _GraphUtil.requires_destroy_resource_sentinel(downstream_boundary, graph):
            # The destroy_resource node is the only leaf node in the DAG.
            # Add a flow control node to the graph that will inherit its state
            # from the nodes that are upstream to the destroy_resource node.
            # This is because Airflow's DagRun final-state logic only looks
            # at leaf node states, and if the destroy_resource step runs
            # successfully after an upstream failure then the DagRun would
            # be marked as a success, which is presumably not the desired
            # behavior.

            if resource.disable_sentinel_node:
                logger.debug(
                    'Not adding a sentinel node for resource %s because '
                    'the DAG or the resource configuration specified '
                    'disable_sentinel_node == True',
                    destroy_resource.name)
            else:
                logger.debug(
                    '`%s` is the only leaf node in the DAG: adding a '
                    'sentinel node to propagate upstream failures to '
                    'the DagRun state',
                    destroy_resource.name)
                fc_node = fc_node_builder(
                    deps=downstream_boundary,
                    name=destroy_resource.name + '-sentinel')
                nodes_to_add.append(fc_node)

        graph.add_nodes_from(nodes_to_add)

        for node in downstream_boundary:
            logger.debug(
                'Adding edge between node %s and %s', node, nodes_to_add)
            graph.add_edges_from((node, new_node) for new_node in nodes_to_add)

    @staticmethod
    def prune_node(node, graph):
        if node not in graph:
            raise Exception('Cannot prune node {}: not present in graph {}'.format(
                node, graph))

        upstream = list(graph.predecessors(node))
        downstream = list(graph.successors(node))

        logger.debug('Pruning node %s with upstream nodes %s and downstream nodes %s',
                     node, upstream, downstream)

        graph.add_edges_from((up, down)
                             for up in upstream for down in downstream)

        graph.remove_node(node)

        return graph

    @staticmethod
    def prune_nodes(graph, nodes=None, node_selector=None):
        if not nodes and not node_selector:
            return graph

        components_before_prune = list(
            nx.algorithms.components.weakly_connected_components(graph))

        prune_nodes = nodes if nodes else \
            [node for node in graph.nodes if node_selector(node)]

        for node in prune_nodes:
            _GraphUtil.prune_node(node, graph)

        components_after_prune = list(
            nx.algorithms.components.weakly_connected_components(graph))

        if len(components_before_prune) < len(components_after_prune):
            logger.warning('Pruning of nodes %s from graph %s caused components to '
                           'become disconnected! Components before prune: %s; after prune: %s',
                           prune_nodes,
                           graph,
                           components_before_prune,
                           components_after_prune)

        return graph


class OperatorGraph(object):
    def get_upstream_surface(self):
        return _GraphUtil.get_upstream_surface(self.graph)

    def get_downstream_surface(self):
        return _GraphUtil.get_downstream_surface(self.graph)

    def upstream_dependency_set(self, node):
        return _GraphUtil.upstream_dependency_set(node, self.graph)

    def downstream_dependency_set(self, node):
        return _GraphUtil.downstream_dependency_set(node, self.graph)

    def ordered(self):
        return _GraphUtil.ordered(self.graph)

    def __init__(self, dag, default_task_args=None):
        self.dag = dag

        self.fc_node_builder = functools.partial(
            util.make_flow_control_node,
            default_task_args=default_task_args or {})

        main_graph = _GraphUtil.build_subgraph(
            dag.get('operators', []) +
            dag.get('sub_dags', []) +
            dag.get('generators', []))

        before_graph = _GraphUtil.build_subgraph(dag.get('before', []))
        after_graph = _GraphUtil.build_subgraph(dag.get('after', []))

        resource_graph = _GraphUtil.build_subgraph(dag.get('resources', []))

        self.graph = nx.algorithms.operators.union_all([
            main_graph,
            before_graph,
            after_graph])

        self.nodes = {node.name: node for node in self.graph.nodes()}
        for upstream_node in _GraphUtil.get_downstream_surface(before_graph):
            for downstream_node in _GraphUtil.get_upstream_surface(main_graph):
                self.graph.add_edge(upstream_node, downstream_node)

        for downstream_node in _GraphUtil.get_upstream_surface(after_graph):
            for upstream_node in _GraphUtil.get_downstream_surface(main_graph):
                self.graph.add_edge(upstream_node, downstream_node)

        # Finally get all resources and figure out how to attach them to the
        # graph
        for resource in resource_graph:
            logger.debug('Attaching resource %s to graph', resource)
            _GraphUtil.attach_create_resource(resource, self.graph, self.fc_node_builder)
            _GraphUtil.attach_destroy_resource(resource, self.graph, self.fc_node_builder)

        self._attach_generator_flow_control()

    def _attach_generator_flow_control(self):
        generators = frozenset(self.dag.get('generators', []))

        for generator in generators:
            downstream_generators = generators & _GraphUtil.downstream_dependency_set(
                generator, self.graph)

            if not downstream_generators:
                continue

            _GraphUtil.attach_flow_control_between(
                upstream=frozenset([generator]),
                downstream=downstream_generators,
                graph=self.graph,
                fc_node_builder=self.fc_node_builder)



import setuptools
import versioneer
import os

import versioneer

extras = {
        'github': [ 'github3.py>=1.1' ],
        }

long_description = None
with open(os.path.join(os.path.dirname(__file__), 'README.md')) as _in:
    long_description = _in.read()

setuptools.setup(
    name = 'boundary-layer',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author = 'Kevin McHale',
    author_email = 'kmchale@etsy.com',
    description = 'Builds Airflow DAGs from configuration files',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/etsy/boundary-layer',
    packages = setuptools.find_packages(),
    package_data = {
        'boundary_layer': [
            'builders/templates/*.j2',
            ],
        'boundary_layer_default_plugin': [
            'config/generators/*.yaml',
            'config/operators/*.yaml',
            'config/resources/*.yaml',
            'config/subdags/*.yaml',
            ]
        },
    entry_points = {
            'boundary_layer_plugins': [
                'default=boundary_layer_default_plugin.plugin:DefaultPlugin',
            ]
        },
    extras_require = extras,
    install_requires = [
        'enum34>=1.1.6,<2.0;python_version<"3.4"',
        'semver>=2.7.0,<3.0',
        'jsonschema>=2.6.0,<3.0',
        'jinja2>=2.8.1,<3.0',
        'pyyaml>=4.2b1',
        'marshmallow>=2.13.6,<3.0',
        'networkx>=2.1,<2.3',
        'xmltodict>=0.11.0,<1.0',
        'six>=1.11.0,<2.0',
    ],
    scripts = ['bin/boundary-layer'],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],
    license = 'Apache License 2.0',
    keywords = 'airflow',
    zip_safe = False,
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
)



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import os
from boundary_layer.plugins import BasePlugin, PluginPriority
from .oozie_plugin import DefaultOozieParserPlugin
from .preprocessors import DateStringToDatetime, BuildTimedelta, EnsureRenderedStringPattern
from .preprocessors import KubernetesPrep, PubsubMessageDataToBinaryString


class DefaultPlugin(BasePlugin):
    name = 'default'

    priority = PluginPriority.DEFAULT

    configs_path = os.path.join(os.path.dirname(__file__), 'config')

    oozie_plugin_cls = DefaultOozieParserPlugin

    property_preprocessors = [
        DateStringToDatetime,
        BuildTimedelta,
        EnsureRenderedStringPattern,
        KubernetesPrep,
        PubsubMessageDataToBinaryString]



# -*- coding: utf-8 -*-
# Copyright 2018 Etsy Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from boundary_layer.builders.base import DagBuilderBase


class SubDagBuilder(DagBuilderBase):
    indent_operators = True

    def preamble(self):
        assert len(self.reference_path) > 1, \
            'Invalid Sub-DAG reference path: {}'.format(self.reference_path)

        template = self.get_jinja_template('subdag_preamble.j2')

        dag_args = {
            'dag_id': self.build_dag_id()
        }

        return template.render(
            subdag_operator_name=self.dag['name'],
            dag_args=dag_args,
            specs=self.specs,
        )

    def epilogue(self):
        template = self.get_jinja_template('subdag_epilogue.j2')

        return template.render()
